"""
The MIT License (MIT)

Copyright (c) 2012, Florian Finkernagel <finkernagel@imt.uni-marburg.de>

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import sys
import pytest
import pypipegraph as ppg
from .shared import write, assertRaises, read, Dummy, append

global_test = 0


@pytest.mark.usefixtures("new_pipegraph")
class TestJobs:
    def test_assert_singletonicity_of_jobs(self, new_pipegraph):
        new_pipegraph.new_pipegraph()
        of = "out/a"
        data_to_write = "hello"

        def do_write():
            write(of, data_to_write)

        job = ppg.FileGeneratingJob(of, do_write)
        job2 = ppg.FileGeneratingJob(of, do_write)
        assert job is job2

    def test_redifining_a_jobid_with_different_class_raises(self, new_pipegraph):
        new_pipegraph.new_pipegraph()
        of = "out/a"
        data_to_write = "hello"

        def do_write():
            write(of, data_to_write)

        ppg.FileGeneratingJob(of, do_write)

        def load():
            return "shu"

        def inner():
            ppg.DataLoadingJob(of, load)

        assertRaises(ValueError, inner)

    def test_addition(self, new_pipegraph):
        def write_func(of):
            def do_write():
                write(of, "do_write done")

            return of, do_write

        new_pipegraph.new_pipegraph()
        jobA = ppg.FileGeneratingJob(*write_func("out/a"))
        jobB = ppg.FileGeneratingJob(*write_func("out/b"))
        jobC = ppg.FileGeneratingJob(*write_func("out/c"))
        jobD = ppg.FileGeneratingJob(*write_func("out/d"))

        aAndB = jobA + jobB
        assert len(aAndB) == 2
        assert jobA in aAndB
        assert jobB in aAndB

        aAndBandC = aAndB + jobC
        assert jobA in aAndBandC
        assert jobB in aAndBandC
        assert jobC in aAndBandC

        aAndBAndD = jobD + aAndB
        assert jobA in aAndBAndD
        assert jobB in aAndBAndD
        assert jobD in aAndBAndD

        cAndD = jobC + jobD
        all = aAndB + cAndD
        assert len(all) == 4
        assert jobA in all
        assert jobB in all
        assert jobC in all
        assert jobD in all

    def test_raises_on_non_str_job_id(self):
        def inner():
            ppg.FileGeneratingJob(1234, lambda: None)

        assertRaises(TypeError, inner)

    def test_equality_is_identity(self, new_pipegraph):
        def write_func(of):
            def do_write():
                write(of, "do_write done")

            return of, do_write

        new_pipegraph.new_pipegraph()
        jobA = ppg.FileGeneratingJob(*write_func("out/a"))
        jobA1 = ppg.FileGeneratingJob(*write_func("out/a"))
        jobB = ppg.FileGeneratingJob(*write_func("out/b"))
        assert jobA is jobA1
        assert jobA == jobA1
        assert not (jobA == jobB)

    def test_has_hash(self, new_pipegraph):
        new_pipegraph.new_pipegraph()
        jobA = ppg.FileGeneratingJob("out/", lambda: None)
        assert hasattr(jobA, "__hash__")


@pytest.mark.usefixtures("new_pipegraph")
class TestJobs2:
    def test_ignore_code_changes_is_ignored(self):
        jobA = ppg.Job("shu")
        jobA.ignore_code_changes()

    def test_load_raises(self):
        jobA = ppg.Job("shu")

        def inner():
            jobA.load()

        assertRaises(ValueError, inner)

    def test_is_in_dependency_chain_direct(self):
        jobA = ppg.Job("A")
        jobB = ppg.Job("B")
        jobA.depends_on(jobB)
        assert jobA.is_in_dependency_chain(jobB, 100)

    def test_is_in_dependency_chain_direct2(self):
        jobA = ppg.Job("A")
        jobs = []
        for x in range(0, 10):
            j = ppg.Job(str(x))
            if jobs:
                j.depends_on(jobs[-1])
            jobs.append(j)
        jobA.depends_on(jobs[-1])
        assert jobA.is_in_dependency_chain(jobs[0], 100)
        assert jobA.is_in_dependency_chain(jobs[0], 10)
        # max_depth reached -> answer with false
        assert not (jobA.is_in_dependency_chain(jobs[0], 5))

    def test_str(self):
        a = ppg.FileGeneratingJob("out/A", lambda: write("out/A", "hello"))
        assert isinstance(str(a), str)

        a = ppg.ParameterInvariant("out/A", "hello")
        assert isinstance(str(a), str)

        a = ppg.JobGeneratingJob("out/Ax", lambda: "hello")
        assert isinstance(str(a), str)


@pytest.mark.usefixtures("new_pipegraph")
class TestFileGeneratingJob:
    def test_basic(self):
        of = "out/a"
        data_to_write = "hello"

        def do_write():
            print("do_write was called")
            write(of, data_to_write)

        job = ppg.FileGeneratingJob(of, do_write)
        job.ignore_code_changes()
        ppg.run_pipegraph()
        assert not (job.failed)
        assert os.path.exists(of)
        op = open(of, "r")
        data = op.read()
        op.close()
        assert data == data_to_write
        assert job.was_run

    def test_cores_needed(self):
        of = "out/a"
        data_to_write = "hello"

        def do_write():
            print("do_write was called")
            write(of, data_to_write)

        job = ppg.FileGeneratingJob(of, do_write)
        job.ignore_code_changes()
        with pytest.raises(ValueError):
            job.cores_needed = 0
        with pytest.raises(ValueError):
            job.cores_needed = "5"
        with pytest.raises(ValueError):
            job.cores_needed = -3
        job.cores_needed = 1
        job.cores_needed = 5
        job.cores_needed = -1
        job.cores_needed = -2

        for i in range(10):
            j = ppg.FileGeneratingJob(
                "out/%i" % i, lambda i=i: write("out/%i" % i, "b")
            )
            if i % 2 == 0:
                j.cores_needed = -2

        ppg.util.global_pipegraph.rc.cores_available = 5
        ppg.run_pipegraph()
        assert not (job.failed)
        assert os.path.exists(of)
        op = open(of, "r")
        data = op.read()
        op.close()
        assert data == data_to_write
        assert job.was_run

    def test_cores_needed2(self):
        # this is meant to trigger the
        # "this job needed to much resources, or was not runnable"
        # case of graph.start_jobs

        for i in range(20):
            j = ppg.FileGeneratingJob(
                "out/%i" % i, lambda i=i: write("out/%i" % i, "b")
            )
            if i % 2 == 0:
                j.cores_needed = 2

        ppg.util.global_pipegraph.rc.cores_available = 3
        ppg.run_pipegraph()

    def test_needing_more_cores_then_available_raises(self):
        j = ppg.FileGeneratingJob("out/A", lambda: write("out/A", "A"))
        j.cores_needed = 50
        with pytest.raises(ppg.RuntimeError):
            ppg.run_pipegraph()
        assert j.exception is None
        assert j.error_reason == "Needed to much memory/cores"

    def test_basic_with_parameter(self):
        data_to_write = "hello"

        def do_write(filename):
            print("do_write was called")
            write(filename, data_to_write)

        job = ppg.FileGeneratingJob("out/a", do_write)
        job.ignore_code_changes()
        ppg.run_pipegraph()
        assert not (job.failed)
        assert os.path.exists("out/a")
        op = open("out/a", "r")
        data = op.read()
        op.close()
        assert data == data_to_write
        assert job.was_run

    def test_simple_filegeneration_with_function_dependency(self):
        of = "out/a"
        data_to_write = "hello"

        def do_write():
            print("do_write was called")
            write(of, data_to_write)

        job = ppg.FileGeneratingJob(of, do_write)
        # job.ignore_code_changes() this would be the magic line to remove the function dependency
        ppg.run_pipegraph()
        assert not (job.failed)
        assert os.path.exists(of)
        op = open(of, "r")
        data = op.read()
        op.close()
        assert data == data_to_write

    def test_filejob_raises_if_no_data_is_written(self):
        of = "out/a"

        def do_write():
            write("out/A", "")

        job = ppg.FileGeneratingJob(of, do_write)

        def inner():
            ppg.run_pipegraph()

        assertRaises(ppg.RuntimeError, inner)
        assert job.failed
        assert isinstance(job.exception, ppg.JobContractError)
        assert not (os.path.exists(of))

    def test_filejob_empty_allowed(self):
        ppg.FileGeneratingJob("out/A", lambda: write("out/A", ""), empty_ok=True)
        ppg.run_pipegraph()
        assert read("out/A") == ""

    def test_filejob_raising_exception_bubbles(self):
        def do():
            raise ValueError("Hello Exception")

        job = ppg.FileGeneratingJob("out/A", do)
        with pytest.raises(ppg.RuntimeError):
            ppg.run_pipegraph()
        assert "Hello Exception" in str(job.exception)

    def test_simple_filegeneration_removes_file_on_exception(self):
        of = "out/a"
        data_to_write = "hello"

        def do_write():
            write(of, data_to_write)
            raise ValueError("shu")

        job = ppg.FileGeneratingJob(of, do_write)
        try:
            ppg.run_pipegraph()
            raise ValueError("should have raised RuntimeError")
        except ppg.RuntimeError:
            pass
        assert job.failed
        assert not (os.path.exists(of))
        assert isinstance(job.exception, ValueError)

    def test_simple_filegeneration_renames_file_on_exception(self):
        of = "out/a"
        data_to_write = "hello"

        def do_write():
            write(of, data_to_write)
            raise ValueError("shu")

        job = ppg.FileGeneratingJob(of, do_write, rename_broken=True)
        try:
            ppg.run_pipegraph()
            raise ValueError("should have raised RuntimeError")
        except ppg.RuntimeError:
            pass
        assert job.failed
        assert not (os.path.exists(of))
        assert os.path.exists(of + ".broken")
        assert isinstance(job.exception, ValueError)

    def test_simple_filegeneration_captures_stdout_stderr(self):
        of = "out/a"
        data_to_write = "hello"

        def do_write():
            op = open(of, "w")
            op.write(data_to_write)
            op.close()
            print("stdout is cool")
            sys.stderr.write("I am stderr")

        job = ppg.FileGeneratingJob(of, do_write)
        ppg.run_pipegraph()
        assert not (job.failed)
        assert os.path.exists(of)
        op = open(of, "r")
        data = op.read()
        op.close()
        assert data == data_to_write
        assert job.stdout == "stdout is cool\n"
        assert job.stderr == "I am stderr"  # no \n here

    def test_filegeneration_does_not_change_mcp(self):
        global global_test
        global_test = 1
        of = "out/a"
        data_to_write = "hello"

        def do_write():
            write(of, data_to_write)
            global global_test
            global_test = 2

        ppg.FileGeneratingJob(of, do_write)
        ppg.run_pipegraph()
        assert global_test == 1

    def test_file_generation_chaining_simple(self):
        ofA = "out/a"

        def writeA():
            write(ofA, "Hello")

        jobA = ppg.FileGeneratingJob(ofA, writeA)
        ofB = "out/b"

        def writeB():
            op = open(ofB, "w")
            ip = open(ofA, "r")
            op.write(ip.read()[::-1])
            op.close()
            ip.close()

        jobB = ppg.FileGeneratingJob(ofB, writeB)
        jobB.depends_on(jobA)
        ppg.run_pipegraph()
        assert read(ofA) == read(ofB)[::-1]

    def test_file_generation_multicore(self):
        # one fork per FileGeneratingJob...
        ofA = "out/a"

        def writeA():
            write(ofA, "%i" % os.getpid())

        ofB = "out/b"

        def writeB():
            write(ofB, "%i" % os.getpid())

        ppg.FileGeneratingJob(ofA, writeA)
        ppg.FileGeneratingJob(ofB, writeB)
        ppg.run_pipegraph()
        assert read(ofA) != read(ofB)

    def test_filegenerating_two_jobs_same_file(self):
        ppg.MultiFileGeneratingJob(["out/A", "out/B"], lambda: write("out/A", "hello"))
        with pytest.raises(ValueError):
            ppg.MultiFileGeneratingJob(
                ["out/A", "out/C"], lambda: write("out/A", "world")
            )
        ppg.FileGeneratingJob("out/C", lambda: write("out/C", "C"))
        with pytest.raises(ValueError):
            ppg.MultiFileGeneratingJob(["out/C"], lambda: write("out/A", "world"))
        ppg.MultiFileGeneratingJob(["out/D", "out/E"], lambda: write("out/A", "world"))
        with pytest.raises(ValueError):
            ppg.FileGeneratingJob("out/D", lambda: write("out/C", "C"))

    def test_invaliding_removes_file(self, new_pipegraph):
        of = "out/a"
        sentinel = "out/b"

        def do_write():
            if os.path.exists(sentinel):
                raise ValueError("second run")
            write(of, "shu")
            write(sentinel, "done")

        job = ppg.FileGeneratingJob(of, do_write, empty_ok=True)
        dep = ppg.ParameterInvariant("my_params", (1,))
        job.depends_on(dep)
        ppg.run_pipegraph()
        assert os.path.exists(of)
        assert os.path.exists(sentinel)

        new_pipegraph.new_pipegraph()
        job = ppg.FileGeneratingJob(of, do_write)
        dep = ppg.ParameterInvariant(
            "my_params", (2,)
        )  # same name ,changed params, job needs to rerun, but explodes...
        job.depends_on(dep)  # on average, half the mistakes are in the tests...
        try:
            ppg.run_pipegraph()
            raise ValueError("Should not have been reached")
        except ppg.RuntimeError:
            pass
        assert not (os.path.exists(of))

    def test_passing_non_function(self):
        def inner():
            ppg.FileGeneratingJob("out/a", "shu")

        assertRaises(ValueError, inner)

    def test_passing_non_string_as_jobid(self):
        def inner():
            ppg.FileGeneratingJob(5, lambda: 1)

        assertRaises(TypeError, inner)

    def test_exceptions_are_preserved(self):
        def shu():
            write("out/A", "A")
            write("out/Ay", "ax")
            raise IndexError("twenty-five")  # just some exception

        jobA = ppg.FileGeneratingJob("out/A", shu)

        def inner():
            ppg.run_pipegraph()

        assertRaises(ppg.RuntimeError, inner)
        print(jobA.exception)
        assert isinstance(jobA.exception, IndexError)
        assert not (
            os.path.exists("out/A")
        )  # should clobber the resulting files in this case - just a double check to test_invaliding_removes_file
        assert read("out/Ay") == "ax"  # but the job did run, right?

    def test_exceptions_are_preserved2(self):
        def shu():
            write("out/A", "A")
            write("out/Ay", "ax")
            raise TypeError("twenty-five")  # just some exception

        jobA = ppg.FileGeneratingJob("out/A", shu)

        def inner():
            ppg.run_pipegraph()

        assertRaises(ppg.RuntimeError, inner)
        print(jobA.exception)
        assert isinstance(jobA.exception, TypeError)
        assert "twenty-five" in str(jobA.exception)
        assert not (
            os.path.exists("out/A")
        )  # should clobber the resulting files in this case - just a double check to test_invaliding_removes_file
        assert read("out/Ay") == "ax"  # but the job did run, right?

    def test_dumping_graph(self):
        ppg.new_pipegraph(
            quiet=True, invariant_status_filename="shu.dat", dump_graph=True
        )
        ppg.FileGeneratingJob("out/A", lambda: write("out/A", "A"))
        ppg.run_pipegraph()
        pid = ppg.util.global_pipegraph.dump_pid
        os.waitpid(pid, 0)
        print(os.listdir("logs"))
        assert os.path.exists("logs/ppg_graph.gml")

    def test_rename_broken(self):
        def do(of):
            write(of, "hello")
            raise ValueError()

        ppg.FileGeneratingJob("out/A", do, rename_broken=True)

        def do():
            write("out/B", "world")
            raise ValueError()

        ppg.FileGeneratingJob("out/B", do, rename_broken=True)

        def do():
            raise ValueError()

        ppg.FileGeneratingJob("out/C", do, rename_broken=True)

        with pytest.raises(ppg.RuntimeError):
            ppg.run_pipegraph()
        assert read("out/A.broken") == "hello"
        assert read("out/B.broken") == "world"
        assert not os.path.exists("out/A")
        assert not os.path.exists("out/B")
        assert not os.path.exists("out/C")
        assert not os.path.exists("out/C.broken")


@pytest.mark.usefixtures("new_pipegraph")
class TestMultiFileGeneratingJob:
    def test_basic(self, new_pipegraph):
        of = ["out/a", "out/b"]

        def do_write():
            for f in of:
                append(f, "shu")

        ppg.MultiFileGeneratingJob(of, do_write)
        ppg.run_pipegraph()
        for f in of:
            assert read(f) == "shu"
        new_pipegraph.new_pipegraph()
        ppg.MultiFileGeneratingJob(of, do_write)
        ppg.run_pipegraph()
        for f in of:
            assert read(f) == "shu"  # ie. job has net been rerun...
        # but if I now delete one...
        os.unlink(of[0])
        new_pipegraph.new_pipegraph()
        ppg.MultiFileGeneratingJob(of, do_write)
        ppg.run_pipegraph()
        assert read(of[0]) == "shu"
        assert (
            read(of[1]) == "shu"
        )  # Since that file was also deleted when MultiFileGeneratingJob was invalidated...

    def test_empty_raises(self):
        of = ["out/a", "out/b"]

        def do_write():
            for f in of:
                append(f, "shu")
            write(f, "")  # one is empty

        ppg.MultiFileGeneratingJob(of, do_write)
        with pytest.raises(ValueError):
            ppg.run_pipegraph()

    def test_empty_ok(self):
        of = ["out/a", "out/b"]

        def do_write():
            for f in of:
                append(f, "shu")
            write(f, "")  # one is empty

        ppg.MultiFileGeneratingJob(of, do_write, empty_ok=True)
        ppg.run_pipegraph()
        for f in of[:-1]:
            assert read(f) == "shu"
        assert read(of[-1]) == ""

    def test_exception_destroys_all_files(self):
        of = ["out/a", "out/b"]

        def do_write():
            for f in of:
                append(f, "shu")
            raise ValueError("explode")

        ppg.MultiFileGeneratingJob(of, do_write)
        try:
            ppg.run_pipegraph()
            raise ValueError("should not be reached")
        except ppg.RuntimeError:
            pass
        for f in of:
            assert not (os.path.exists(f))

    def test_exception_destroys_renames_files(self):
        of = ["out/a", "out/b"]

        def do_write():
            for f in of:
                append(f, "shu")
            raise ValueError("explode")

        ppg.MultiFileGeneratingJob(of, do_write, rename_broken=True)
        try:
            ppg.run_pipegraph()
            raise ValueError("should not be reached")
        except ppg.RuntimeError:
            pass
        for f in of:
            assert os.path.exists(f + ".broken")

    def test_invalidation_removes_all_files(self, new_pipegraph):
        of = ["out/a", "out/b"]
        sentinel = (
            "out/sentinel"
        )  # hack so this one does something different the second time around...

        def do_write():
            if os.path.exists(sentinel):
                raise ValueError("explode")
            write(sentinel, "shu")
            for f in of:
                append(f, "shu")

        ppg.MultiFileGeneratingJob(of, do_write).depends_on(
            ppg.ParameterInvariant("myparam", (1,))
        )
        ppg.run_pipegraph()
        for f in of:
            assert os.path.exists(f)
        new_pipegraph.new_pipegraph()
        ppg.MultiFileGeneratingJob(of, do_write).depends_on(
            ppg.ParameterInvariant("myparam", (2,))
        )
        try:
            ppg.run_pipegraph()  # since this should blow up
            raise ValueError("should not be reached")
        except ppg.RuntimeError:
            pass
        for f in of:
            assert not (os.path.exists(f))

    def test_passing_not_a_list_of_str(self):
        def inner():
            ppg.MultiFileGeneratingJob("out/a", lambda: 1)

        assertRaises(ValueError, inner)

    def test_passing_non_function(self):
        def inner():
            ppg.MultiFileGeneratingJob(["out/a"], "shu")

        assertRaises(ValueError, inner)

    def test_exceptions_are_preserved(self):
        def shu():
            write("out/A", "A")
            write("out/B", "B")
            write("out/Az", "ax")
            raise IndexError("twenty-five")  # just some exception

        jobA = ppg.MultiFileGeneratingJob(["out/A", "out/B"], shu)

        def inner():
            ppg.run_pipegraph()

        assertRaises(ppg.RuntimeError, inner)
        assert isinstance(jobA.exception, IndexError)
        assert not (
            os.path.exists("out/A")
        )  # should clobber the resulting files in this case - just a double check to test_invaliding_removes_file
        assert not (
            os.path.exists("out/B")
        )  # should clobber the resulting files in this case - just a double check to test_invaliding_removes_file
        assert read("out/Az") == "ax"  # but the job did run, right?

    def raises_on_non_string_filnames(self):
        def inner():
            ppg.MultiFileGeneratingJob(["one", 2], lambda: write("out/A"))

        assertRaises(ValueError, inner)

    def test_raises_on_collision(self):
        def inner():
            ppg.MultiFileGeneratingJob(["test1", "test2"], lambda: 5)
            ppg.MultiFileGeneratingJob(["test2", "test3"], lambda: 5)

        assertRaises(ValueError, inner)

    def test_duplicate_prevention(self):
        param = "A"
        ppg.FileGeneratingJob("out/A", lambda: write("out/A", param))

        def inner():
            ppg.MultiFileGeneratingJob(["out/A"], lambda: write("out/A", param))

        assertRaises(ValueError, inner)

    def test_non_str(self):
        param = "A"

        def inner():
            ppg.MultiFileGeneratingJob([25], lambda: write("out/A", param))

        assertRaises(TypeError, inner)

    def test_non_iterable(self):
        param = "A"
        try:
            ppg.MultiFileGeneratingJob(25, lambda: write("out/A", param))
            assert not ("Exception not raised")
        except TypeError as e:
            print(e)
            assert "filenames was not iterable" in str(e)

    def test_single_stre(self):
        param = "A"

        def inner():
            ppg.MultiFileGeneratingJob("A", lambda: write("out/A", param))

        assertRaises(ValueError, inner)

    def test_can_run_now_case_1(self):
        class HonestIsDoneJob(ppg.FileGeneratingJob):
            def is_done(self, depth=0):
                # we have to force it, otherwise
                # the caching of is_done and stat calls
                # will prevent this case from happening
                return os.path.exists(self.job_id)

        ppg.util.global_pipegraph.rc.cores_available = 2

        def a():
            write("out/A", "A")
            import time

            time.sleep(1)

        a = HonestIsDoneJob("out/A", a)
        b = ppg.FileGeneratingJob("out/B", lambda: write("out/B", "B"))
        c = ppg.FileGeneratingJob("out/C", lambda: write("out/C", "C"))
        a.ignore_code_changes()
        b.ignore_code_changes()
        c.ignore_code_changes()
        c.depends_on(a)
        c.depends_on(b)
        ppg.run_pipegraph()
        assert read("out/C") == "C"


test_modifies_shared_global = []
shared_value = ""


@pytest.mark.usefixtures("new_pipegraph")
class TestDataLoadingJob:
    def test_modifies_slave(self):
        # global shared
        # shared = "I was the the global in the mcp"
        def load():
            test_modifies_shared_global.append("shared data")

        of = "out/a"

        def do_write():
            write(
                of, "\n".join(test_modifies_shared_global)
            )  # this might actually be a problem when defining this?

        dlJo = ppg.DataLoadingJob("myjob", load)
        writejob = ppg.FileGeneratingJob(of, do_write)
        writejob.depends_on(dlJo)

        writejob2 = ppg.FileGeneratingJob(
            "out/b",
            lambda: write("out/b", "b" + "\n".join(test_modifies_shared_global)),
        )
        writejob2.depends_on(dlJo)
        ppg.run_pipegraph()
        assert read(of) == "shared data"
        assert read("out/b") == "bshared data"

    def test_global_statement_works(self):
        # this currently does not work in the cloudpickle transmitted jobs -
        # two jobs refereing to global have different globals afterwards
        # or the 'global shared' does not work as expected after loading
        global shared_value
        shared_value = "I was the the global in the mcp"

        def load():
            global shared_value
            shared_value = "shared data"

        of = "out/a"

        def do_write():
            write(of, shared_value)

        dlJo = ppg.DataLoadingJob("myjob", load)
        writejob = ppg.FileGeneratingJob(of, do_write)
        writejob.depends_on(dlJo)
        ppg.run_pipegraph()
        assert read(of) == "shared data"

    def test_does_not_get_run_without_dep_job(self):
        of = "out/shu"

        def load():
            write(
                of, "shu"
            )  # not the fine english way, but we need a sideeffect that's checkable

        ppg.DataLoadingJob("myjob", load)
        ppg.run_pipegraph()
        assert not (os.path.exists(of))

    def test_does_not_get_run_in_chain_without_final_dep(self):
        of = "out/shu"

        def load():
            write(
                of, "shu"
            )  # not the fine english way, but we need a sideeffect that's checkable

        job = ppg.DataLoadingJob("myjob", load)
        ofB = "out/sha"

        def loadB():
            write(ofB, "sha")

        ppg.DataLoadingJob("myjobB", loadB).depends_on(job)
        ppg.run_pipegraph()
        assert not (os.path.exists(of))
        assert not (os.path.exists(ofB))

    def test_does_get_run_in_chain_all(self):
        of = "out/shu"

        def load():
            write(
                of, "shu"
            )  # not the fine english way, but we need a sideeffect that's checkable

        job = ppg.DataLoadingJob("myjob", load)
        ofB = "out/sha"

        def loadB():
            write(ofB, "sha")

        jobB = ppg.DataLoadingJob("myjobB", loadB).depends_on(job)
        ofC = "out/c"

        def do_write():
            write(ofC, ofC)

        ppg.FileGeneratingJob(ofC, do_write).depends_on(jobB)
        ppg.run_pipegraph()
        assert os.path.exists(of)
        assert os.path.exists(ofB)
        assert os.path.exists(ofC)

    def test_chain_with_filegenerating_works(self):
        of = "out/a"

        def do_write():
            write(of, of)

        jobA = ppg.FileGeneratingJob(of, do_write)
        o = Dummy()

        def do_load():
            o.a = read(of)

        jobB = ppg.DataLoadingJob("loadme", do_load).depends_on(jobA)
        ofB = "out/b"

        def write2():
            write(ofB, o.a)

        ppg.FileGeneratingJob(ofB, write2).depends_on(jobB)
        ppg.run_pipegraph()
        assert read(of) == of
        assert read(ofB) == of

    def test_does_get_run_depending_on_jobgenjob(self):
        of = "out/shu"

        def load():
            write(
                of, "shu"
            )  # not the fine english way, but we need a sideeffect that's checkable

        job = ppg.DataLoadingJob("myjob", load)

        def gen():
            ofB = "out/b"

            def do_write():
                write(ofB, "hello")

            ppg.FileGeneratingJob(ofB, do_write)

        ppg.JobGeneratingJob("mygen", gen).depends_on(job)
        ppg.run_pipegraph()
        assert os.path.exists(of)  # so the data loading job was run
        assert read("out/b") == "hello"  # and so was the jobgen and filegen job.

    def test_passing_non_function(self):
        def inner():
            ppg.DataLoadingJob("out/a", "shu")

        assertRaises(ValueError, inner)

    def test_passing_non_string_as_jobid(self):
        def inner():
            ppg.DataLoadingJob(5, lambda: 1)

        assertRaises(TypeError, inner)

    def test_failing_dataloading_jobs(self):
        o = Dummy()
        of = "out/A"

        def write():
            write(of, o.a)

        def load():
            o.a = "shu"
            raise ValueError()

        job_fg = ppg.FileGeneratingJob(of, write)
        job_dl = ppg.DataLoadingJob("doload", load)
        job_fg.depends_on(job_dl)
        try:
            ppg.run_pipegraph()
            raise ValueError("should not be reached")
        except ppg.RuntimeError:
            pass
        assert not (os.path.exists(of))
        assert job_dl.failed
        assert job_fg.failed
        assert isinstance(job_dl.exception, ValueError)

    def test_prev_dataloading_jobs_not_done_if_there_is_a_non_dataloading_job_inbetween_that_is_done(
        self
    ):
        # so, A = DataLoadingJob, B = FileGeneratingJob, C = DataLoadingJob, D = FileGeneratingJob
        # D.depends_on(C)
        # C.depends_on(B)
        # B.depends_on(A)
        # B is done.
        # D is not
        # since a would be loaded, and then cleaned up right away (because B is Done)
        # it should never be loaded
        o = Dummy()

        def a():
            o.a = "A"
            append("out/A", "A")

        def b():
            append("out/B", "B")
            append("out/Bx", "B")

        def c():
            o.c = "C"
            append("out/C", "C")

        def d():
            append("out/D", "D")
            append("out/Dx", "D")

        jobA = ppg.DataLoadingJob("out/A", a)
        jobB = ppg.FileGeneratingJob("out/B", b)
        jobC = ppg.DataLoadingJob("out/C", c)
        jobD = ppg.FileGeneratingJob("out/D", d)
        jobD.depends_on(jobC)
        jobC.depends_on(jobB)
        jobB.depends_on(jobA)
        jobA.ignore_code_changes()
        jobB.ignore_code_changes()
        jobC.ignore_code_changes()
        jobD.ignore_code_changes()
        write("out/B", "already done")
        assert not (os.path.exists("out/D"))
        ppg.run_pipegraph()
        assert read("out/D") == "D"
        assert read("out/Dx") == "D"
        assert not (
            os.path.exists("out/A")
        )  # A was not executed (as per the premise of the test)
        assert not (
            os.path.exists("out/Bx")
        )  # so B was not executed (we removed the function invariants for this test)
        assert read("out/C") == "C"

    def test_sending_a_non_pickable_exception_data_loading(self):
        class UnpickableException(Exception):
            def __getstate__(self):
                raise ValueError("Can't pickle me")

        def load():
            raise UnpickableException()

        jobA = ppg.DataLoadingJob("out/A", load)
        jobB = ppg.FileGeneratingJob("out/B", lambda: True)
        jobB.depends_on(jobA)

        def inner():
            ppg.run_pipegraph()

        assertRaises(ppg.RuntimeError, inner)
        assert isinstance(jobA.exception, str)

    def test_sending_a_non_pickable_exception_file_generating(self):
        class UnpickableException(Exception):
            def __getstate__(self):
                raise ValueError("Can't pickle me")

        def load():
            raise UnpickableException()

        jobB = ppg.FileGeneratingJob("out/B", load)

        with pytest.raises(ppg.RuntimeError):
            ppg.run_pipegraph()
        assert isinstance(jobB.exception, str)

    def test_creating_jobs_in_file_generating_are_ignored(self):
        def load():
            ppg.util.global_pipegraph.new_jobs = (
                {}
            )  # just to see if we can reach the check in the resource coordinator!
            c = ppg.FileGeneratingJob("out/C", lambda: write("out/C", "C"))
            write("out/A", "A")
            return [c]

        ppg.FileGeneratingJob("out/A", load)
        ppg.run_pipegraph()
        assert read("out/A") == "A"
        assert not os.path.exists("out/C")

    def test_creating_jobs_in_data_loading(self):
        def load():
            ppg.FileGeneratingJob("out/C", lambda: write("out/C", "C"))

        a = ppg.FileGeneratingJob("out/A", lambda: write("out/A", "A"))
        b = ppg.DataLoadingJob("out/B", load)
        a.depends_on(b)
        with pytest.raises(ppg.RuntimeError):
            ppg.run_pipegraph()
        assert isinstance(b.exception, ppg.JobContractError)
        assert (
            "Trying to add new jobs to running pipeline without having new_jobs "
            in str(b.exception)
        )

    def test_job_returning_value_without_modifying_jobgraph(self):
        class BrokenJob(ppg.FileGeneratingJob):
            def run(self):
                return 55

        a = BrokenJob("out/A", lambda: None)
        with pytest.raises(ppg.RuntimeError):
            ppg.run_pipegraph()
        assert isinstance(a.exception, ppg.JobContractError)
        assert "Job returned a value " in str(a.exception)


@pytest.mark.usefixtures("new_pipegraph")
class TestAttributeJob:
    def test_basic_attribute_loading(self):
        o = Dummy()

        def load():
            return "shu"

        job = ppg.AttributeLoadingJob("load_dummy_shu", o, "a", load)
        of = "out/a"

        def do_write():
            write(of, o.a)

        ppg.FileGeneratingJob(of, do_write).depends_on(job)
        ppg.run_pipegraph()
        assert read(of) == "shu"

    def test_chained(self):
        o = Dummy()

        def load():
            return "shu"

        def load2():
            return o.a + "sha"

        job = ppg.AttributeLoadingJob("load_dummy_shu", o, "a", load)
        job2 = ppg.AttributeLoadingJob("load_dummy_shu2", o, "b", load2)
        of = "out/a"

        def do_write():
            write(of, o.b)

        ppg.FileGeneratingJob(of, do_write).depends_on(job2)
        job2.depends_on(job)
        ppg.run_pipegraph()
        assert read(of) == "shusha"

    def test_attribute_loading_does_not_affect_mcp(self):
        o = Dummy()

        def load():
            return "shu"

        job = ppg.AttributeLoadingJob("load_dummy_shu", o, "a", load)
        of = "out/a"

        def do_write():
            write(of, o.a)

        ppg.FileGeneratingJob(of, do_write).depends_on(job)
        ppg.run_pipegraph()
        assert read(of) == "shu"
        assert not (hasattr(o, "a"))

    def test_attribute_loading_does_not_run_withot_dependency(self):
        o = Dummy()
        tf = "out/testfile"

        def load():
            write(tf, "hello")
            return "shu"

        ppg.AttributeLoadingJob("load_dummy_shu", o, "a", load)
        ppg.run_pipegraph()
        assert not (hasattr(o, "a"))
        assert not (os.path.exists(tf))

    def test_attribute_disappears_after_direct_dependency(self):
        o = Dummy()
        job = ppg.AttributeLoadingJob("load_dummy_shu", o, "a", lambda: "shu")
        of = "out/A"

        def do_write():
            write(of, o.a)

        fgjob = ppg.FileGeneratingJob(of, do_write).depends_on(job)
        of2 = "out/B"

        def later_write():
            write(of2, o.a)

        ppg.FileGeneratingJob(of2, later_write).depends_on(fgjob)
        try:
            ppg.run_pipegraph()
            raise ValueError("should not be reached")
        except ppg.RuntimeError:
            pass
        assert read(of) == "shu"
        assert not (os.path.exists(of2))

    def test_attribute_disappears_after_direct_dependencies(self):
        o = Dummy()
        job = ppg.AttributeLoadingJob("load_dummy_shu", o, "a", lambda: "shu")
        of = "out/A"

        def do_write():
            write(of, o.a)

        fgjob = ppg.FileGeneratingJob(of, do_write).depends_on(job)
        of2 = "out/B"

        def later_write():
            write(of2, o.a)

        fgjobB = ppg.FileGeneratingJob(of2, later_write).depends_on(
            fgjob
        )  # now, this one does not depend on job, o it should not be able to access o.a
        of3 = "out/C"

        def also_write():
            write(of3, o.a)

        fgjobC = ppg.FileGeneratingJob(of3, also_write).depends_on(job)
        fgjobB.depends_on(
            fgjobC
        )  # otherwise, B might be started C returned, and the cleanup will not have occured!
        try:
            ppg.run_pipegraph()
            raise ValueError("should not be reached")
        except ppg.RuntimeError:
            pass
        assert read(of) == "shu"
        assert read(of3) == "shu"
        assert not (os.path.exists(of2))

    def test_passing_non_string_as_attribute(self):
        o = Dummy()

        def inner():
            ppg.AttributeLoadingJob("out/a", o, 5, 55)

        assertRaises(ValueError, inner)

    def test_passing_non_function(self):
        o = Dummy()

        def inner():
            ppg.AttributeLoadingJob("out/a", o, "a", 55)

        assertRaises(ValueError, inner)

    def test_passing_non_string_as_jobid(self):
        o = Dummy()

        def inner():
            ppg.AttributeLoadingJob(5, o, "a", lambda: 55)

        assertRaises(TypeError, inner)

    def test_no_swapping_attributes_for_one_job(self):
        def cache():
            return list(range(0, 100))

        o = Dummy()
        ppg.AttributeLoadingJob("out/A", o, "a", cache)

        def inner():
            ppg.AttributeLoadingJob("out/A", o, "b", cache)

        assertRaises(ppg.JobContractError, inner)

    def test_raises_on_non_string_attribute_name(self):
        def inner():
            o = Dummy()
            ppg.AttributeLoadingJob("out/A", o, 23, lambda: 5)

        assertRaises(ValueError, inner)

    def test_raises_on_non_function_callback(self):
        def inner():
            o = Dummy()
            ppg.AttributeLoadingJob("out/A", o, 23, 55)

        assertRaises(ValueError, inner)

    def test_no_swapping_objects_for_one_job(self):
        def cache():
            return list(range(0, 100))

        o = Dummy()
        o2 = Dummy()
        ppg.CachedAttributeLoadingJob("out/A", o, "a", cache)

        def inner():
            ppg.CachedAttributeLoadingJob("out/A", o2, "a", cache)

        assertRaises(ppg.JobContractError, inner)

    def test_ignore_code_changes(self, new_pipegraph):
        def a():
            append("out/Aa", "A")
            return "5"

        o = Dummy()
        jobA = ppg.AttributeLoadingJob("out/A", o, "a", a)
        jobA.ignore_code_changes()
        jobB = ppg.FileGeneratingJob("out/B", lambda: write("out/B", o.a))
        jobB.depends_on(jobA)
        ppg.run_pipegraph()
        assert read("out/Aa") == "A"
        assert read("out/B") == "5"
        new_pipegraph.new_pipegraph()

        def b():
            append("out/Aa", "B")
            return "5"

        jobA = ppg.AttributeLoadingJob("out/A", o, "a", b)
        jobA.ignore_code_changes()
        jobB = ppg.FileGeneratingJob("out/B", lambda: write("out/B", o.a))
        jobB.depends_on(jobA)
        ppg.run_pipegraph()
        # not rerun
        assert read("out/Aa") == "A"
        assert read("out/B") == "5"

    def test_callback_must_be_callable(self):
        def inner():
            o = Dummy()
            ppg.AttributeLoadingJob("load_dummy_shu", o, "a", "shu")

        assertRaises(ValueError, inner)


@pytest.mark.usefixtures("new_pipegraph")
class TestTempFileGeneratingJob:
    def test_basic(self):
        temp_file = "out/temp"

        def write_temp():
            write(temp_file, "hello")

        temp_job = ppg.TempFileGeneratingJob(temp_file, write_temp)
        assert temp_job.is_temp_job
        ofA = "out/A"

        def write_A():
            write(ofA, read(temp_file))

        fgjob = ppg.FileGeneratingJob(ofA, write_A)
        fgjob.depends_on(temp_job)
        ppg.run_pipegraph()
        assert read(ofA) == "hello"
        assert not (os.path.exists(temp_file))

    def test_does_not_get_return_if_output_is_done(self, new_pipegraph):
        temp_file = "out/temp"
        out_file = "out/A"
        count_file = "out/count"
        normal_count_file = "out/countA"

        def write_count():
            try:
                count = read(out_file)
                count = count[: count.find(":")]
            except IOError:
                count = "0"
            count = int(count) + 1
            write(out_file, str(count) + ":" + read(temp_file))
            append(normal_count_file, "A")

        def write_temp():
            write(temp_file, "temp")
            append(count_file, "X")

        jobA = ppg.FileGeneratingJob(out_file, write_count)
        jobTemp = ppg.TempFileGeneratingJob(temp_file, write_temp)
        jobA.depends_on(jobTemp)
        ppg.run_pipegraph()
        assert not (os.path.exists(temp_file))
        assert read(out_file) == "1:temp"
        assert read(count_file) == "X"
        assert read(normal_count_file) == "A"
        # now, rerun. Tempfile has been deleted,
        # and should not be regenerated
        new_pipegraph.new_pipegraph()
        jobA = ppg.FileGeneratingJob(out_file, write_count)
        jobTemp = ppg.TempFileGeneratingJob(temp_file, write_temp)
        jobA.depends_on(jobTemp)
        ppg.run_pipegraph()
        assert not (os.path.exists(temp_file))
        assert read(out_file) == "1:temp"
        assert read(count_file) == "X"
        assert read(normal_count_file) == "A"

    def test_does_not_get_return_if_output_is_not(self, new_pipegraph):
        temp_file = "out/temp"
        out_file = "out/A"
        count_file = "out/count"
        normal_count_file = "out/countA"

        def write_count():
            try:
                count = read(out_file)
                count = count[: count.find(":")]
            except IOError:
                count = "0"
            count = int(count) + 1
            write(out_file, str(count) + ":" + read(temp_file))
            append(normal_count_file, "A")

        def write_temp():
            write(temp_file, "temp")
            append(count_file, "X")

        jobA = ppg.FileGeneratingJob(out_file, write_count)
        jobTemp = ppg.TempFileGeneratingJob(temp_file, write_temp)
        jobA.depends_on(jobTemp)
        ppg.run_pipegraph()
        assert not (os.path.exists(temp_file))
        assert read(out_file) == "1:temp"
        assert read(count_file) == "X"
        assert read(normal_count_file) == "A"
        # now, rerun. Tempfile has been deleted,
        # and should  be regenerated
        os.unlink(out_file)
        new_pipegraph.new_pipegraph()
        jobA = ppg.FileGeneratingJob(out_file, write_count)
        jobTemp = ppg.TempFileGeneratingJob(temp_file, write_temp)
        jobA.depends_on(jobTemp)
        ppg.run_pipegraph()
        assert not (os.path.exists(temp_file))
        assert read(out_file) == "1:temp"  # since the outfile was removed...
        assert read(count_file) == "XX"
        assert read(normal_count_file) == "AA"

    def test_dependand_explodes(self, new_pipegraph):
        temp_file = "out/temp"

        def write_temp():
            append(temp_file, "hello")

        temp_job = ppg.TempFileGeneratingJob(temp_file, write_temp)
        ofA = "out/A"

        def write_A():
            raise ValueError("shu")

        fgjob = ppg.FileGeneratingJob(ofA, write_A)
        fgjob.depends_on(temp_job)
        try:
            ppg.run_pipegraph()
            raise ValueError("should not be reached")
        except ppg.RuntimeError:
            pass
        # ppg.run_pipegraph()
        assert not (os.path.exists(ofA))
        assert os.path.exists(temp_file)

        new_pipegraph.new_pipegraph()

        def write_A_ok():
            write(ofA, read(temp_file))

        temp_job = ppg.TempFileGeneratingJob(temp_file, write_temp)
        fgjob = ppg.FileGeneratingJob(ofA, write_A_ok)
        fgjob.depends_on(temp_job)
        ppg.run_pipegraph()

        assert read(ofA) == "hello"  # tempfile job has not been rerun
        assert not (os.path.exists(temp_file))  # and the tempfile has been removed...

    def test_removes_tempfile_on_exception(self):
        temp_file = "out/temp"

        def write_temp():
            write(temp_file, "hello")
            raise ValueError("should")

        temp_job = ppg.TempFileGeneratingJob(temp_file, write_temp)
        ofA = "out/A"

        def write_A():
            write(ofA, read(temp_file))

        fgjob = ppg.FileGeneratingJob(ofA, write_A)
        fgjob.depends_on(temp_job)
        try:
            ppg.run_pipegraph()
            raise ValueError("should not be reached")
        except ppg.RuntimeError:
            pass
        assert not (os.path.exists(temp_file))
        assert not (os.path.exists(ofA))

    def test_renames_tempfile_on_exception_if_requested(self):
        temp_file = "out/temp"

        def write_temp():
            write(temp_file, "hello")
            raise ValueError("should")

        temp_job = ppg.TempFileGeneratingJob(temp_file, write_temp, rename_broken=True)
        ofA = "out/A"

        def write_A():
            write(ofA, read(temp_file))

        fgjob = ppg.FileGeneratingJob(ofA, write_A)
        fgjob.depends_on(temp_job)
        try:
            ppg.run_pipegraph()
            raise ValueError("should not be reached")
        except ppg.RuntimeError:
            pass
        assert not (os.path.exists(temp_file))
        assert os.path.exists(temp_file + ".broken")
        assert not (os.path.exists(ofA))

    def test_passing_non_function(self):
        def inner():
            ppg.TempFileGeneratingJob("out/a", "shu")

        assertRaises(ValueError, inner)

    def test_passing_non_string_as_jobid(self):
        def inner():
            ppg.TempFileGeneratingJob(5, lambda: 1)

        assertRaises(TypeError, inner)

    def test_rerun_because_of_new_dependency_does_not_rerun_old(self, new_pipegraph):
        jobA = ppg.FileGeneratingJob(
            "out/A", lambda: append("out/A", read("out/temp")) or append("out/Ab", "A")
        )
        jobB = ppg.TempFileGeneratingJob("out/temp", lambda: write("out/temp", "T"))
        jobA.depends_on(jobB)
        ppg.run_pipegraph()
        assert not (os.path.exists("out/temp"))
        assert read("out/A") == "T"
        assert read("out/Ab") == "A"  # ran once

        new_pipegraph.new_pipegraph()
        jobA = ppg.FileGeneratingJob("out/A", lambda: append("out/A", read("out/temp")))
        jobB = ppg.TempFileGeneratingJob("out/temp", lambda: write("out/temp", "T"))
        jobA.depends_on(jobB)
        jobC = ppg.FileGeneratingJob("out/C", lambda: append("out/C", read("out/temp")))
        jobC.depends_on(jobB)
        ppg.run_pipegraph()
        assert not (os.path.exists("out/temp"))
        assert read("out/Ab") == "A"  # ran once, not rewritten
        assert read("out/C") == "T"  # a new file

    def test_chaining_multiple(self):
        jobA = ppg.TempFileGeneratingJob("out/A", lambda: write("out/A", "A"))
        jobB = ppg.TempFileGeneratingJob(
            "out/B", lambda: write("out/B", read("out/A") + "B")
        )
        jobC = ppg.TempFileGeneratingJob(
            "out/C", lambda: write("out/C", read("out/A") + "C")
        )
        jobD = ppg.FileGeneratingJob(
            "out/D", lambda: write("out/D", read("out/B") + read("out/C"))
        )
        jobD.depends_on(jobC)
        jobD.depends_on(jobB)
        jobC.depends_on(jobA)
        jobB.depends_on(jobA)
        ppg.run_pipegraph()
        assert read("out/D") == "ABAC"
        assert not (os.path.exists("out/A"))
        assert not (os.path.exists("out/B"))
        assert not (os.path.exists("out/C"))

    def test_chaining_multiple_differently(self):
        jobA = ppg.TempFileGeneratingJob("out/A", lambda: write("out/A", "A"))
        jobB = ppg.TempFileGeneratingJob(
            "out/B", lambda: write("out/B", read("out/A") + "B")
        )
        jobD = ppg.FileGeneratingJob(
            "out/D", lambda: write("out/D", read("out/B") + "D")
        )
        jobE = ppg.FileGeneratingJob(
            "out/E", lambda: write("out/E", read("out/B") + "E")
        )
        jobF = ppg.FileGeneratingJob(
            "out/F", lambda: write("out/F", read("out/A") + "F")
        )
        jobD.depends_on(jobB)
        jobE.depends_on(jobB)
        jobB.depends_on(jobA)
        jobF.depends_on(jobA)
        ppg.run_pipegraph()
        assert read("out/D") == "ABD"
        assert read("out/E") == "ABE"
        assert read("out/F") == "AF"
        assert not (os.path.exists("out/A"))
        assert not (os.path.exists("out/B"))
        assert not (os.path.exists("out/C"))

    def test_rerun_because_of_new_dependency_does_not_rerun_old_chained(
        self, new_pipegraph
    ):
        jobA = ppg.TempFileGeneratingJob("out/A", lambda: write("out/A", "A"))
        jobB = ppg.TempFileGeneratingJob(
            "out/B", lambda: write("out/B", read("out/A") + "B")
        )
        jobC = ppg.FileGeneratingJob(
            "out/C",
            lambda: write("out/C", read("out/B") + "C") or append("out/Cx", "1"),
        )
        jobB.depends_on(jobA)
        jobC.depends_on(jobB)
        ppg.run_pipegraph()
        assert read("out/C") == "ABC"
        assert read("out/Cx") == "1"

        new_pipegraph.new_pipegraph()
        jobA = ppg.TempFileGeneratingJob("out/A", lambda: write("out/A", "A"))
        jobB = ppg.TempFileGeneratingJob(
            "out/B", lambda: write("out/B", read("out/A") + "B")
        )
        jobC = ppg.FileGeneratingJob(
            "out/C",
            lambda: write("out/C", read("out/B") + "C") or append("out/Cx", "1"),
        )
        jobD = ppg.FileGeneratingJob(
            "out/D",
            lambda: write("out/D", read("out/A") + "D") or append("out/Dx", "1"),
        )
        jobB.depends_on(jobA)
        jobC.depends_on(jobB)
        jobD.depends_on(jobA)
        ppg.run_pipegraph()
        assert read("out/D") == "AD"
        assert read("out/Dx") == "1"
        assert read("out/C") == "ABC"
        assert read("out/Cx") == "1"

        new_pipegraph.new_pipegraph()
        jobA = ppg.TempFileGeneratingJob(
            "out/A", lambda: write("out/A", "a")
        )  # note changing function code!
        jobB = ppg.TempFileGeneratingJob(
            "out/B", lambda: write("out/B", read("out/A") + "B")
        )
        jobC = ppg.FileGeneratingJob(
            "out/C",
            lambda: write("out/C", read("out/B") + "C") or append("out/Cx", "1"),
        )
        jobD = ppg.FileGeneratingJob(
            "out/D",
            lambda: write("out/D", read("out/A") + "D") or append("out/Dx", "1"),
        )
        jobB.depends_on(jobA)
        jobC.depends_on(jobB)
        jobD.depends_on(jobA)
        ppg.run_pipegraph()
        assert read("out/D") == "aD"
        assert read("out/Dx") == "11"  # both get rerun
        assert read("out/C") == "aBC"
        assert read("out/Cx") == "11"

    def test_cleanup_if_never_run(self, new_pipegraph):
        temp_file = "out/temp"

        def write_temp():
            write(temp_file, "hello")

        def write_a():
            write("A", "hello")

        temp_job = ppg.TempFileGeneratingJob(temp_file, write_temp)
        temp_job.ignore_code_changes()
        jobA = ppg.FileGeneratingJob("A", write_a)
        jobA.ignore_code_changes()
        write_a()  # so the file is there!
        ppg.run_pipegraph()
        assert not (os.path.exists("out/temp"))
        new_pipegraph.new_pipegraph()
        write_temp()
        assert os.path.exists("out/temp")
        temp_job = ppg.TempFileGeneratingJob(temp_file, write_temp)
        temp_job.ignore_code_changes()
        temp_job.do_cleanup_if_was_never_run = True
        ppg.run_pipegraph()
        assert not (os.path.exists("out/temp"))


@pytest.mark.usefixtures("new_pipegraph")
class TestMultiTempFileGeneratingJob:
    def test_basic(self):
        temp_files = ["out/temp", "out/temp2"]

        def write_temp():
            for temp_file in temp_files:
                write(temp_file, "hello")

        temp_job = ppg.MultiTempFileGeneratingJob(temp_files, write_temp)
        assert temp_job.is_temp_job
        ofA = "out/A"

        def write_A():
            write(ofA, read(temp_files[0]) + read(temp_files[1]))

        fgjob = ppg.FileGeneratingJob(ofA, write_A)
        fgjob.depends_on(temp_job)
        ppg.run_pipegraph()
        assert read(ofA) == "hellohello"
        assert not (os.path.exists(temp_files[0]))
        assert not (os.path.exists(temp_files[1]))

    def test_basic_dependes_were_done(self):
        temp_files = ["out/temp", "out/temp2"]

        def write_temp():
            write("temp_sentinel", "one")
            for temp_file in temp_files:
                write(temp_file, "hello")

        temp_job = ppg.MultiTempFileGeneratingJob(temp_files, write_temp)
        temp_job.ignore_code_changes()
        assert temp_job.is_temp_job
        ofA = "out/A"

        def write_A():
            write(ofA, read(temp_files[0]) + read(temp_files[1]))

        fgjob = ppg.FileGeneratingJob(ofA, write_A)
        fgjob.ignore_code_changes()
        fgjob.depends_on(temp_job)
        write(ofA, "two")
        ppg.run_pipegraph()
        assert read(ofA) == "two"
        assert not os.path.exists("temp_sentinel")

    def raises_on_non_string_filnames(self):
        def inner():
            ppg.MultiTempFileGeneratingJob(["one", 2], lambda: write("out/A"))

        assertRaises(ValueError, inner)

    def test_raises_on_collision(self):
        def inner():
            ppg.MultiTempFileGeneratingJob(["test1", "test2"], lambda: 5)
            ppg.MultiTempFileGeneratingJob(["test2", "test3"], lambda: 5)

        assertRaises(ValueError, inner)

    def test_duplicate_prevention(self):
        param = "A"
        ppg.FileGeneratingJob("out/A", lambda: write("out/A", param))

        def inner():
            ppg.MultiTempFileGeneratingJob(["out/A"], lambda: write("out/A", param))

        assertRaises(ValueError, inner)

    def test_non_str(self):
        param = "A"

        def inner():
            ppg.MultiTempFileGeneratingJob([25], lambda: write("out/A", param))

        assertRaises(TypeError, inner)

    def test_non_iterable(self):
        param = "A"
        try:
            ppg.MultiTempFileGeneratingJob(25, lambda: write("out/A", param))
            assert not ("Exception not raised")
        except TypeError as e:
            print(e)
            assert "filenames was not iterable" in str(e)


@pytest.mark.usefixtures("new_pipegraph")
class TestTempFilePlusGeneratingJob:
    def test_basic(self):
        ppg.new_pipegraph(quiet=False, dump_graph=False)
        temp_file = "out/temp"
        keep_file = "out/keep"

        def write_temp():
            write(temp_file, "hello")
            write(keep_file, "hello")

        temp_job = ppg.TempFilePlusGeneratingJob(temp_file, keep_file, write_temp)
        assert temp_job.is_temp_job
        ofA = "out/A"

        def write_A():
            write(ofA, read(temp_file))

        fgjob = ppg.FileGeneratingJob(ofA, write_A)
        fgjob.depends_on(temp_job)
        ppg.run_pipegraph()
        assert read(ofA) == "hello"
        assert not (os.path.exists(temp_file))
        assert os.path.exists(keep_file)

    def test_not_run_if_dependands_all_done(self):
        ppg.new_pipegraph(quiet=False, dump_graph=False)
        temp_file = "out/temp"
        keep_file = "out/keep"

        def write_temp():
            write(temp_file, "hello")
            write(keep_file, "hello")

        temp_job = ppg.TempFilePlusGeneratingJob(temp_file, keep_file, write_temp)
        temp_job.ignore_code_changes()
        write(keep_file, "keepme")
        ofA = "out/A"

        def write_A():
            write(ofA, read(temp_file))

        fgjob = ppg.FileGeneratingJob(ofA, write_A)
        fgjob.ignore_code_changes()
        write(ofA, "two")
        fgjob.depends_on(temp_job)
        ppg.run_pipegraph()
        assert read(ofA) == "two"
        assert not (os.path.exists(temp_file))
        assert read(keep_file) == "keepme"

    def test_raises_on_keep_equal_temp_file(self):
        temp_file = "out/temp"
        keep_file = temp_file

        def write_temp():
            write(temp_file, "hello")
            write(keep_file, "hello")

        def inner():
            ppg.TempFilePlusGeneratingJob(temp_file, keep_file, write_temp)

        assertRaises(ValueError, inner)

    def test_does_not_get_return_if_output_is_done(self, new_pipegraph):
        temp_file = "out/temp"
        keep_file = "out/keep"
        out_file = "out/A"
        count_file = "out/count"
        normal_count_file = "out/countA"

        def write_count():
            try:
                count = read(out_file)
                count = count[: count.find(":")]
            except IOError:
                count = "0"
            count = int(count) + 1
            write(out_file, str(count) + ":" + read(temp_file))
            append(normal_count_file, "A")

        def write_temp():
            write(temp_file, "temp")
            write(keep_file, "temp")
            append(count_file, "X")

        jobA = ppg.FileGeneratingJob(out_file, write_count)
        jobTemp = ppg.TempFilePlusGeneratingJob(temp_file, keep_file, write_temp)
        jobA.depends_on(jobTemp)
        ppg.run_pipegraph()
        assert not (os.path.exists(temp_file))
        assert read(out_file) == "1:temp"
        assert read(count_file) == "X"
        assert read(normal_count_file) == "A"
        assert os.path.exists(keep_file)
        # now, rerun. Tempfile has been deleted,
        # and should not be regenerated
        new_pipegraph.new_pipegraph()
        jobA = ppg.FileGeneratingJob(out_file, write_count)
        jobTemp = ppg.TempFilePlusGeneratingJob(temp_file, keep_file, write_temp)
        jobA.depends_on(jobTemp)
        ppg.run_pipegraph()
        assert not (os.path.exists(temp_file))
        assert read(out_file) == "1:temp"
        assert read(count_file) == "X"
        assert read(normal_count_file) == "A"
        assert os.path.exists(keep_file)

    def test_fails_if_keep_file_is_not_generated(self):
        temp_file = "out/temp"
        keep_file = "out/keep"
        out_file = "out/A"
        count_file = "out/count"
        normal_count_file = "out/countA"

        def write_count():
            try:
                count = read(out_file)
                count = count[: count.find(":")]
            except IOError:
                count = "0"
            count = int(count) + 1
            write(out_file, str(count) + ":" + read(temp_file))
            append(normal_count_file, "A")

        def write_temp():
            write(temp_file, "temp")
            # write(keep_file, 'temp')
            append(count_file, "X")

        jobA = ppg.FileGeneratingJob(out_file, write_count)
        jobTemp = ppg.TempFilePlusGeneratingJob(temp_file, keep_file, write_temp)
        jobA.depends_on(jobTemp)

        def inner():
            ppg.run_pipegraph()

        assertRaises(ppg.RuntimeError, inner)
        assert not (os.path.exists(out_file))
        assert not (os.path.exists(keep_file))
        assert os.path.exists(temp_file)

    def test_does_get_rerun_if_keep_file_is_gone(self, new_pipegraph):
        temp_file = "out/temp"
        keep_file = "out/keep"
        out_file = "out/A"
        count_file = "out/count"
        normal_count_file = "out/countA"

        def write_count():
            try:
                count = read(out_file)
                count = count[: count.find(":")]
            except IOError:
                count = "0"
            count = int(count) + 1
            write(out_file, str(count) + ":" + read(temp_file))
            append(normal_count_file, "A")

        def write_temp():
            write(temp_file, "temp")
            write(keep_file, "temp")
            append(count_file, "X")

        jobA = ppg.FileGeneratingJob(out_file, write_count)
        jobTemp = ppg.TempFilePlusGeneratingJob(temp_file, keep_file, write_temp)
        jobA.depends_on(jobTemp)
        ppg.run_pipegraph()
        assert not (os.path.exists(temp_file))
        assert read(out_file) == "1:temp"
        assert read(count_file) == "X"
        assert read(normal_count_file) == "A"
        assert os.path.exists(keep_file)
        os.unlink(keep_file)
        new_pipegraph.new_pipegraph()
        jobA = ppg.FileGeneratingJob(out_file, write_count)
        jobTemp = ppg.TempFilePlusGeneratingJob(temp_file, keep_file, write_temp)
        jobA.depends_on(jobTemp)
        ppg.run_pipegraph()
        assert read(out_file) == "1:temp"
        assert read(count_file) == "XX"  # where we see the temp file job ran again
        assert read(normal_count_file) == "AA"  # which is where we see it ran again...
        assert os.path.exists(keep_file)


@pytest.mark.usefixtures("new_pipegraph")
class TestFinalJobs:
    def test_correct_dependencies(self):
        o = Dummy()

        def a():
            o.a = "A"
            append("out/A", "A")

        def b():
            append("out/B", "B")
            append("out/Bx", "B")

        def c():
            o.c = "C"
            append("out/C", "C")

        def d():
            append("out/D", "D")
            append("out/Dx", "D")

        jobA = ppg.DataLoadingJob("out/A", a)
        jobB = ppg.FileGeneratingJob("out/B", b)
        jobC = ppg.DataLoadingJob("out/C", c)
        jobD = ppg.FileGeneratingJob("out/D", d)
        # jobD.depends_on(jobC)
        # jobC.depends_on(jobB)
        jobB.depends_on(jobA)
        final_job = ppg.FinalJob(
            "da_final",
            lambda: write(
                "out/final", "done" + read("out/A") + read("out/B") + read("out/D")
            ),
        )
        ppg.util.global_pipegraph.connect_graph()
        print(final_job.prerequisites)
        for x in jobB, jobC, jobD:
            assert x in final_job.prerequisites
            assert final_job in x.dependants
        assert not (jobA in final_job.prerequisites)
        ppg.run_pipegraph()
        assert read("out/final") == "doneABD"
        assert not os.path.exists(
            "out/C"
        )  # dataloading job does not get run just because of FinalJob

    def test_cannot_depend_on_final_job(self):
        jobA = ppg.FileGeneratingJob("A", lambda: write("A", "A"))
        final_job = ppg.FinalJob("da_final", lambda: None)
        try:
            jobA.depends_on(final_job)
            assert not ("Exception not raised")
        except ppg.JobContractError as e:
            print(e)
            assert "No jobs can depend on FinalJobs" in str(e)

    def test_final_job_depends_on_raises(self):
        jobA = ppg.FileGeneratingJob("A", lambda: write("A", "A"))
        final_job = ppg.FinalJob("da_final", lambda: None)
        try:
            final_job.depends_on(jobA)
            assert not ("Exception not raised")
        except ppg.JobContractError as e:
            print(e)
            assert "Final jobs can not have explicit dependencies " in str(e)


@pytest.mark.usefixtures("new_pipegraph")
class TestJobList:
    def raises_on_non_job(self):
        def inner():
            ppg.JobList("shu")

        assertRaises(ValueError, inner)

    def test_raises_on_non_job_in_list(self):
        with pytest.raises(ValueError):
            ppg.JobList([ppg.ParameterInvariant("a", 23), "shu"])

    def test_add(self):
        jobA = ppg.FileGeneratingJob("A", lambda: write("A", "A"))
        jobB = ppg.FileGeneratingJob("B", lambda: write("A", "A"))
        jobC = ppg.FileGeneratingJob("C", lambda: write("C", "A"))
        jobD = ppg.FileGeneratingJob("D", lambda: write("D", "A"))
        l1 = ppg.JobList([jobA])
        l2 = l1 + jobB
        assert len(l2) == 2
        l2 = l1 + jobB
        assert len(l2) == 2
        l3 = l2 + ppg.JobList(jobC)
        assert len(l3) == 3
        l4 = l3 + [jobD]
        assert len(l4) == 4

    def test_depends_on(self):
        jobA = ppg.FileGeneratingJob("A", lambda: write("A", "A"))
        jobB = ppg.FileGeneratingJob("B", lambda: write("A", "A"))
        jobC = ppg.FileGeneratingJob("C", lambda: write("C", "A"))
        l1 = ppg.JobList([jobA, jobB])
        l1.depends_on(jobC)
        assert jobC in jobA.prerequisites
        assert jobC in jobB.prerequisites
        ppg.util.global_pipegraph.connect_graph()
        assert jobA in jobC.dependants
        assert jobB in jobC.dependants

    def test_depends_on_list(self):
        jobA = ppg.FileGeneratingJob("A", lambda: write("A", "A"))
        jobB = ppg.FileGeneratingJob("B", lambda: write("A", "A"))
        jobC = ppg.FileGeneratingJob("C", lambda: write("C", "A"))
        l1 = ppg.JobList([jobC])
        l1.depends_on([jobA, jobB])
        assert jobA in jobC.prerequisites
        assert jobB in jobC.prerequisites
        ppg.util.global_pipegraph.connect_graph()
        assert jobC in jobA.dependants
        assert jobC in jobB.dependants

    def test_depends_on_args(self):
        jobA = ppg.FileGeneratingJob("A", lambda: write("A", "A"))
        jobB = ppg.FileGeneratingJob("B", lambda: write("A", "A"))
        jobC = ppg.FileGeneratingJob("C", lambda: write("C", "A"))
        l1 = ppg.JobList([jobC])
        l1.depends_on(jobA, jobB)
        assert jobA in jobC.prerequisites
        assert jobB in jobC.prerequisites
        ppg.util.global_pipegraph.connect_graph()
        assert jobC in jobA.dependants
        assert jobC in jobB.dependants

    def test_str(self):
        jobA = ppg.FileGeneratingJob("A", lambda: write("A", "A"))
        l1 = ppg.JobList([jobA])
        x = str(l1)
        assert x.startswith("JobList of ")

    def test_adding_two_jobs(self):
        jobA = ppg.FileGeneratingJob("A", lambda: write("A", "A"))
        jobB = ppg.FileGeneratingJob("B", lambda: write("A", "A"))
        x = jobA + jobB
        assert isinstance(x, ppg.JobList)
        assert len(x) == 2


@pytest.mark.usefixtures("new_pipegraph")
class TestNotebookJobs:
    def write_notebook(self, filename, payload):
        with open(filename, "w") as op:
            op.write(
                """{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "with open(\\"out/A\\", \\"w\\") as op:\\n",
    "    op.write(\\"%s\\")\\n",
    "with open(\\"out/tag\\", \\"a+\\") as op:\\n",
    "    op.write(\\"x\\")\\n"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.6.6"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}"""
                % (payload,)
            )

    def test_notebook(self, new_pipegraph):
        self.write_notebook("out/test.ipynb", "hello")
        ppg.NotebookJob("out/test.ipynb", False)
        ppg.run_pipegraph()
        assert read("out/A") == "hello"
        assert read("out/tag") == "x"

        new_pipegraph.new_pipegraph()
        ppg.NotebookJob("out/test.ipynb", False)
        ppg.run_pipegraph()
        assert read("out/A") == "hello"
        assert read("out/tag") == "x"  # no rerun

        new_pipegraph.new_pipegraph()
        self.write_notebook("out/test.ipynb", "world")
        ppg.NotebookJob("out/test.ipynb", False)
        ppg.run_pipegraph()
        assert read("out/tag") == "xx"  # rerun
        assert read("out/A") == "world"

        new_pipegraph.new_pipegraph()
        ppg.NotebookJob("out/test.ipynb", True)
        ppg.run_pipegraph()
        assert read("out/A") == "world"
        assert read("out/tag") == "xx"  # no rerun

        new_pipegraph.new_pipegraph()
        ppg.NotebookJob("out/test.ipynb", True)
        ppg.MultiFileGeneratingJob(["world"], lambda: write("world", "two"))
        ppg.run_pipegraph()
        assert read("out/A") == "world"
        assert read("out/tag") == "xxx"  # no rerun

    def test_func_returning_same_job(self):
        def get_job():
            def load():
                return self.id

            return ppg.MultiFileGeneratingJob(["shu"], load)

        a = get_job()
        b = get_job()
        assert a is b

    def test_func_returning_different(self):
        first = [True]

        def get_job():
            def load():
                return self.id

            def load2():
                return self.id

            ppg.MultiFileGeneratingJob(["shu"], load if first[0] else load2)
            first[0] = False

        get_job()
        with pytest.raises(ValueError):
            get_job()



"""
The MIT License (MIT)

Copyright (c) 2012, Florian Finkernagel <finkernagel@imt.uni-marburg.de>

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import subprocess
import pytest
import pypipegraph as ppg
from .shared import write, read, append, Dummy, assertRaises


@pytest.mark.usefixtures("new_pipegraph")
class TestResourceCoordinator:
    def test_jobs_that_need_all_cores_are_spawned_one_by_one(self):
        # we'll determine this by the start respective end times..
        ppg.new_pipegraph(
            ppg.resource_coordinators.LocalSystem(max_cores_to_use=2),
            quiet=True,
            dump_graph=False,
        )
        jobA = ppg.FileGeneratingJob("out/A", lambda: write("out/A", "A"))
        jobB = ppg.FileGeneratingJob("out/B", lambda: write("out/B", "B"))
        jobA.cores_needed = -1
        jobB.cores_needed = -1
        ppg.run_pipegraph()
        assert read("out/A") == "A"
        assert read("out/B") == "B"
        if jobA.start_time < jobB.start_time:
            first_job = jobA
            second_job = jobB
        else:
            first_job = jobB
            second_job = jobA
        print(
            "times",
            first_job.start_time,
            first_job.stop_time,
            second_job.start_time,
            second_job.stop_time,
        )
        if jobA.start_time is None:
            raise ValueError("JobA did not run")
        assert first_job.stop_time < second_job.start_time

    def test_jobs_concurrent_jobs_run_concurrently(self):
        # we'll determine this by the start respective end times..
        ppg.new_pipegraph(
            ppg.resource_coordinators.LocalSystem(max_cores_to_use=2),
            quiet=True,
            dump_graph=False,
        )
        jobA = ppg.FileGeneratingJob("out/A", lambda: write("out/A", "A"))
        jobB = ppg.FileGeneratingJob("out/B", lambda: write("out/B", "B"))
        jobA.cores_needed = 1
        jobB.cores_needed = 1
        ppg.run_pipegraph()
        assert read("out/A") == "A"
        assert read("out/B") == "B"
        if jobA.start_time < jobB.start_time:
            first_job = jobA
            second_job = jobB
        else:
            first_job = jobB
            second_job = jobA
        print(
            "times",
            first_job.start_time,
            first_job.stop_time,
            second_job.start_time,
            second_job.stop_time,
        )
        if jobA.start_time is None:
            raise ValueError("JobA did not run")
        assert first_job.stop_time > second_job.start_time


class CantDepickle:
    """A class that can't be depickled (throws a type error,
    just like the numpy.maskedarray does occacionally)"""

    def __getstate__(self):
        return {"shu": "5"}

    def __setstate__(self, state):
        print(state)
        raise TypeError("I can be pickled, but not unpickled")


@pytest.mark.usefixtures("new_pipegraph")
class TestingTheUnexpectedTests:
    def test_job_killing_python(self):
        def dies():
            import sys

            # logging.info("Now terminating child python")
            sys.exit(5)

        fg = ppg.FileGeneratingJob("out/A", dies)
        try:
            ppg.util.global_pipegraph.rc.timeout = 1
            ppg.run_pipegraph()
            raise ValueError("should not be reached")
        except ppg.RuntimeError:
            pass
        assert not (os.path.exists("out/A"))
        assert isinstance(fg.exception, ppg.JobDiedException)
        assert fg.exception.exit_code == 5

    def test_job_killing_python_stdout_stderr_logged(self):
        def dies():
            import sys

            # logging.info("Now terminating child python")
            print("hello")
            sys.stderr.write("I am stderr\n")
            sys.stdout.flush()
            sys.exit(5)

        fg = ppg.FileGeneratingJob("out/A", dies)
        try:
            ppg.util.global_pipegraph.rc.timeout = 1
            ppg.run_pipegraph()
            raise ValueError("should not be reached")
        except ppg.RuntimeError:
            pass
        assert not (os.path.exists("out/A"))
        assert isinstance(fg.exception, ppg.JobDiedException)
        assert fg.exception.exit_code == 5
        assert fg.stdout == "hello\n"
        assert fg.stderr == "I am stderr\n"

    def test_unpickle_bug_prevents_single_job_from_unpickling(self):
        def do_a():
            write("out/A", "A")
            append("out/As", "A")

        ppg.FileGeneratingJob("out/A", do_a)

        def do_b():
            write("out/B", "A")
            append("out/Bs", "A")

        job_B = ppg.FileGeneratingJob("out/B", do_b)
        cd = CantDepickle()
        job_parameter_unpickle_problem = ppg.ParameterInvariant("C", (cd,))
        job_B.depends_on(job_parameter_unpickle_problem)
        ppg.run_pipegraph()
        assert read("out/A") == "A"
        assert read("out/As") == "A"
        assert read("out/B") == "A"
        assert read("out/Bs") == "A"
        print("second run")
        ppg.new_pipegraph(dump_graph=False)

        ppg.FileGeneratingJob("out/A", do_a)
        job_B = ppg.FileGeneratingJob("out/B", do_b)
        job_parameter_unpickle_problem = ppg.ParameterInvariant("C", (cd,))
        job_B.depends_on(job_parameter_unpickle_problem)
        with pytest.raises(ppg.RuntimeError):
            ppg.run_pipegraph()
        assert read("out/A") == "A"
        assert read("out/As") == "A"
        assert read("out/B") == "A"
        assert (
            read("out/Bs") == "AA"
        )  # this one got rerun because we could not load the invariant...

    def testing_import_does_not_hang(self):  # see python issue22853
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(__file__))
        p = subprocess.Popen(
            ["python", "_import_does_not_hang.py"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = p.communicate()
        print(stdout, stderr)
        assert b"OK" in stdout
        os.chdir(old_dir)

    def test_older_jobs_added_back_to_new_pipegraph(self, new_pipegraph):
        a = ppg.FileGeneratingJob("out/A", lambda of: write(of, "a"))
        ppg.util.global_pipegraph.run()
        new_pipegraph.new_pipegraph()
        b = ppg.FileGeneratingJob("out/B", lambda of: write(of, "b"))
        with pytest.raises(ppg.PyPipeGraphError):
            a.depends_on(b)
        with pytest.raises(ppg.PyPipeGraphError):
            b.depends_on(a)


class TestsNotImplemented:
    @pytest.mark.xfail
    def test_temp_jobs_and_gen_jobs(self):
        # DependencyInjection A creates TempJob B and job C (c is already done)
        # DependencyInjeciton D (dep on A) creates TempJob B and job E
        # Now, When A runs, B is created, and not added to the jobs-to-run list
        # since it is not necessary (with C being done).
        # now D runs, B would not be required by E, but does not get added to the
        # run list (since it is not new), and later on, the sanity check crashes.

        # alternativly, if C is not done, execution order is A, B, C. Then cleanup
        # for B happens, then D is run, the E explodes, because cleanup has been done!

        # now, if A returns B, it get's injected into the dependenies of D,
        # the exeuction order is correct, but B get's done no matter what because D
        # now requires it, even if both C and E have already been done.

        # what a conundrum
        raise NotImplementedError()

    @pytest.mark.xfail
    def test_cached_job_done_but_gets_invalidated_by_dependency_injection_generated_job(
        self
    ):
        # very similar to the previous case,
        # this basically directly get's you into the 'Job execution order territory...'
        raise NotImplementedError()


@pytest.mark.usefixtures("new_pipegraph")
class TestPathLib:
    def test_multifilegenerating_job_requires_string_filenames(self):
        import pathlib

        x = lambda: 5  # noqa:E731
        ppg.MultiFileGeneratingJob(["a"], x)
        ppg.MultiFileGeneratingJob([pathlib.Path("a")], x)

        def inner():
            ppg.MultiFileGeneratingJob([0])

        assertRaises(TypeError, inner)

        def inner():
            ppg.MultiFileGeneratingJob([b"a"])  # bytes is not a string type

        assertRaises(TypeError, inner)

    def test_accepts(self):
        import pathlib

        write("aaa", "hello")
        write("bbb", "hello")
        write("ccc", "hello")
        a = ppg.FileTimeInvariant(pathlib.Path("aaa"))
        a1 = ppg.MultiFileInvariant([pathlib.Path("bbb"), "ccc"])
        b = ppg.FileGeneratingJob(
            pathlib.Path("b"),
            lambda of: write(of, "bb" + read("aaa") + read("bbb") + read("ccc")),
        )
        b.depends_on(a)
        b.depends_on(a1)

        dd = Dummy()

        def mf():
            write("c", "cc" + read("g"))
            write("d", "dd" + read("h") + dd.attr)
            write("e", "ee" + read("i") + read("j"))

        c = ppg.MultiFileGeneratingJob([pathlib.Path("c"), "d", pathlib.Path("e")], mf)
        c.depends_on(b)
        d = ppg.FunctionInvariant(pathlib.Path("f"), lambda x: x + 1)
        c.depends_on(d)
        e = ppg.ParameterInvariant(pathlib.Path("c"), "hello")
        c.depends_on(e)
        f = ppg.TempFileGeneratingJob(pathlib.Path("g"), lambda: write("g", "gg"))
        c.depends_on(f)

        def tmf():
            write("h", "hh")
            write("i", "ii")

        g = ppg.MultiTempFileGeneratingJob([pathlib.Path("h"), "i"], tmf)
        c.depends_on(g)

        def tpf():
            write("j", "jjjj")
            write("k", "kkkk")

        h = ppg.TempFilePlusGeneratingJob(pathlib.Path("j"), pathlib.Path("k"), tpf)
        c.depends_on(h)

        i = ppg.CachedDataLoadingJob(
            pathlib.Path("l"), lambda: write("l", "llll"), lambda res: res
        )
        c.depends_on(i)

        m = ppg.CachedAttributeLoadingJob(pathlib.Path("m"), dd, "attr", lambda: "55")
        c.depends_on(m)
        ppg.run_pipegraph()
        assert read("aaa") == "hello"
        assert read("b") == "bbhellohellohello"
        assert read("c") == "ccgg"
        assert read("d") == "ddhh55"
        assert read("e") == "eeiijjjj"
        assert not (os.path.exists("g"))
        assert not (os.path.exists("h"))
        assert not (os.path.exists("i"))
        assert not (os.path.exists("j"))
        assert read("k") == "kkkk"


def test_fixture_without_class(new_pipegraph):
    import pathlib

    assert "run/.test_fixture_without_class" in str(pathlib.Path(".").absolute())


def test_job_or_filename(new_pipegraph):
    a, dep_a = ppg.util.job_or_filename("out/A")
    assert a == "out/A"
    assert len(dep_a) == 1
    assert isinstance(dep_a[0], ppg.RobustFileChecksumInvariant)
    j = ppg.FileGeneratingJob("out/B", lambda: None)
    b, dep_b = ppg.util.job_or_filename(j)
    assert b == "out/B"
    assert dep_b[0] is j
    assert len(dep_b) == 1

    c, dep_c = ppg.util.job_or_filename(None)
    assert c is None
    assert not dep_c


def test_stat_cache(new_pipegraph):
    import time

    write("out/A", "A")
    assert ppg.util.stat("out/A")
    os.unlink("out/A")
    assert ppg.util.stat("out/A")  # cached...
    time.sleep(1)
    with pytest.raises(FileNotFoundError):
        ppg.util.stat("out/A")  # cache invalidated


def test_interactive_import(new_pipegraph):
    # just so at least the import part of interactive is under coverage
    import pypipegraph.interactive  # noqa:F401


def test_logging(new_pipegraph):
    import logging
    my_logger = logging.getLogger('pypipegraph')
    h = logging.FileHandler(filename='ppg.log', mode='w')
    my_logger.addHandler(h)
    logging.getLogger().warning("Should not be in the log.")
    try:
        my_logger.setLevel(logging.DEBUG)
        f = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        h.setFormatter(f)
        ppg.FileGeneratingJob('out/A', lambda: write('out/A', 'A'))
        ppg.run_pipegraph()
    finally:
        my_logger.removeHandler(h)
    assert os.path.exists('ppg.log')
    d = read('ppg.log')
    assert not ('Should not be in the log.\n' in d)
    assert 'pypipegraph - INFO' in d
    assert 'pypipegraph - DEBUG' in d



# -*- coding: utf-8 -*-
from pkg_resources import get_distribution, DistributionNotFound

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:  # pragma: no cover
    __version__ = "unknown"
finally:
    del get_distribution, DistributionNotFound

from .graph import run_pipegraph, new_pipegraph
from .ppg_exceptions import (
    RuntimeError,
    RuntimeException,
    CycleError,
    JobContractError,
    PyPipeGraphError,
    JobDiedException,
    NothingChanged,
)
from . import util

from .job import (
    Job,
    JobList,
    FileGeneratingJob,
    MultiFileGeneratingJob,
    DataLoadingJob,
    AttributeLoadingJob,
    TempFileGeneratingJob,
    TempFilePlusGeneratingJob,
    CachedAttributeLoadingJob,
    CachedDataLoadingJob,
    PlotJob,
    CombinedPlotJob,
    FunctionInvariant,
    ParameterInvariant,
    FileTimeInvariant,
    FileChecksumInvariant,
    RobustFileChecksumInvariant,
    FileInvariant,
    MultiFileInvariant,
    JobGeneratingJob,
    DependencyInjectionJob,
    FinalJob,
    MemMappedDataLoadingJob,
    MultiTempFileGeneratingJob,
    NotebookJob,
)

assert_uniqueness_of_object = util.assert_uniqueness_of_object

all = [
    run_pipegraph,
    new_pipegraph,
    RuntimeError,
    RuntimeException,
    CycleError,
    JobContractError,
    PyPipeGraphError,
    JobDiedException,
    NothingChanged,
    Job,
    JobList,
    FileGeneratingJob,
    MultiFileGeneratingJob,
    DataLoadingJob,
    AttributeLoadingJob,
    TempFileGeneratingJob,
    TempFilePlusGeneratingJob,
    CachedAttributeLoadingJob,
    CachedDataLoadingJob,
    PlotJob,
    CombinedPlotJob,
    FunctionInvariant,
    ParameterInvariant,
    FileTimeInvariant,
    FileChecksumInvariant,
    RobustFileChecksumInvariant,
    MultiFileInvariant,
    FileInvariant,
    JobGeneratingJob,
    DependencyInjectionJob,
    FinalJob,
    MemMappedDataLoadingJob,
    util,
    MultiTempFileGeneratingJob,
    NotebookJob,
]



import pytest


def read(filename):
    """simply read a file"""
    op = open(filename)
    data = op.read()
    op.close()
    return data


def write(filename, string):
    """open file for writing, dump string, close file"""
    op = open(filename, "w")
    op.write(string)
    op.close()


def append(filename, string):
    """open file for appending, dump string, close file"""
    op = open(filename, "a")
    op.write(string)
    op.close()


def writeappend(filename_write, filename_append, string):
    write(filename_write, string)
    append(filename_append, string)


def assertRaises(exception, func):
    with pytest.raises(exception):
        func()


class Dummy(object):
    pass



import pypipegraph
import urllib2

# This creates a global Pipegraph object
# All new jobs will automatically register with it.
pypipegraph.new_pipegraph()

url = "http://code.google.com/p/pypipegraph"  # what website to download.
target_file = "website.html"  # where to store the downloaded website
output_filename = "result.tab"  # where to store the final counts


# now, we need a function that downloads from url and stores to target_file
def download():
    request = urllib2.urlopen(url)
    data = request.read()
    request.close()
    file_handle = open(target_file, "wb")
    file_handle.write(data)
    file_handle.close()


# and we turn it into a FileGeneratingJob
# which receives two parameters: the filename, and the function to call
# function will be called without parameters.
# Note how we pass in the function download, not it's result download()
job_download = pypipegraph.FileGeneratingJob(target_file, download)


# read the downloaded file, count characters, and write to output_filename
def count_characters_and_write_output():
    file_handle = open(target_file, "rb")
    data = file_handle.read()
    count = len(data[data.find("<body") :])
    file_handle.close()
    file_handle = open(output_filename, "wb")
    file_handle.write("%s\t%i\n" % (url, count))
    file_handle.close()


# this also becomes a FileGeneratingJob
job_count = pypipegraph.FileGeneratingJob(
    output_filename, count_characters_and_write_output
)
# now, job_count needs target_file, which is created by job_download, and we need to tell it so
job_count.depends_on(job_download)

# and submit the go on
pypipegraph.run_pipegraph()
# first job_download will be run by calling download(), then job_count will be executed.
# You'll see lines like 'Done 0 of 2 jobs, 1 running', followed by 'Pipegraph done. Executed 4 jobs. 0 failed.



import pytest
from pathlib import Path
import shutil
import os
import pypipegraph as ppg


# support code to remove test created files
# only if the test suceeded


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"

    setattr(item, "rep_" + rep.when, rep)


@pytest.fixture
def new_pipegraph(request):
    if request.cls is None:
        target_path = Path(request.fspath).parent / "run" / ("." + request.node.name)
    else:
        target_path = (
            Path(request.fspath).parent
            / "run"
            / (request.cls.__name__ + "." + request.node.name)
        )
    if target_path.exists():  # pragma: no cover
        shutil.rmtree(target_path)
    target_path = target_path.absolute()
    old_dir = Path(os.getcwd()).absolute()
    try:
        first = [False]

        def np():
            if not first[0]:
                Path(target_path).mkdir(parents=True, exist_ok=True)
                os.chdir(target_path)
                Path("logs").mkdir()
                Path("cache").mkdir()
                Path("results").mkdir()
                Path("out").mkdir()
                import logging

                h = logging.getLogger("pypipegraph")
                h.setLevel(logging.WARNING)
                first[0] = True

            rc = ppg.resource_coordinators.LocalSystem(1)
            ppg.new_pipegraph(rc, quiet=True, dump_graph=False)
            ppg.util.global_pipegraph.result_dir = Path("results")
            g = ppg.util.global_pipegraph
            g.new_pipegraph = np
            return g

        def finalize():
            if hasattr(request.node, "rep_setup"):

                if request.node.rep_setup.passed and (
                    request.node.rep_call.passed
                    or request.node.rep_call.outcome == "skipped"
                ):
                    try:
                        shutil.rmtree(target_path)
                    except OSError:  # pragma: no cover
                        pass

        request.addfinalizer(finalize)
        yield np()

    finally:
        os.chdir(old_dir)



import pypipegraph
import urllib2
import hashlib

pypipegraph.new_pipegraph()

output_filename = "result.tab"  # where to store the final counts


# each call to download_job will return a job that downloads just this url.
def download_job(url):
    target_file = (
        "website_%s" % hashlib.md5(url).hexdigest()
    )  # we need a unique name for each target file.

    def do_download():
        request = urllib2.urlopen(url)
        data = request.read()
        request.close()
        file_handle = open(target_file, "wb")
        file_handle.write(data)
        file_handle.close()

    return pypipegraph.FileGeneratingJob(target_file, do_download)


def retrieve_urls():
    # now I said we were downloading these, but to make the tutorial independand,
    # we'll fake that bit, ok?
    # just pretend ;).
    return [
        "http://code.google.com/p/pypipegraph",
        "http://code.google.com/p/pypipegraph/w/list",
    ]


# this function will be executed after pypipegraph.run_pipegraph and returns
# a list of jobs that job_count will then depend on.
urls = None


def generate_download_jobs():
    jobs = []
    global urls  # we use a global to communicate with the later job
    # please note that this is only possible in Job generating and data loading jobs,
    # not in the output jobs.
    urls = retrieve_urls()
    for url in urls:
        jobs.append(download_job(url))
    jobs.append(  # this makes sure that if you remove urls, the output job would also rerun.
        # adding urls not seen before would make it rerun either way.
        pypipegraph.ParameterInvariant("retrieved_urls", urls)
    )
    return jobs


# A DependencyInjectionJob allows us to make job_count depend on the jobs returned by generate_download_jobs
# during the runtime of the graph
job_generate = pypipegraph.DependencyInjectionJob(
    "retrieve_url", generate_download_jobs
)


# now, this needs to be more fancy as well.
def count_characters_and_write_output():
    counts = {}
    for url in urls:
        dj = download_job(
            url
        )  # this will get us the *same* object as download_job(url) returned the first time
        target_file = (
            dj.job_id
        )  # the job_id for FileGeneratingJob is the output filename.

        file_handle = open(target_file, "rb")
        data = file_handle.read()
        count = len(data)
        counts[url] = len(data)
        file_handle.close()
    file_handle = open(output_filename, "wb")
    for url, count in counts.items():
        file_handle.write("%s\t%i\n" % (url, count))
    file_handle.close()


# this also becomes a FileGeneratingJob
job_count = pypipegraph.FileGeneratingJob(
    output_filename, count_characters_and_write_output
)
# and we can tell it to depend on all jobs at once!
job_count.depends_on(job_generate)

# and submit the go on
pypipegraph.run_pipegraph()



if __name__ == "__main__":
    modules_to_test = ["test_pypipegraph", "test_plotjobs"]

    import os

    os.environ["PYPIPEGRAPH_DO_COVERAGE"] = os.path.abspath(
        os.path.join(os.path.dirname(__file__), ".coveragerc")
    )
    import coverage

    with open(".coveragerc", "wb") as op:
        op.write(
            b"""
[run]
data_file = %s
parallel=True

[report]
include = *pypipegraph*
        """
            % (
                os.path.join(
                    os.path.dirname(os.path.abspath(__file__)), ".coverageX"
                ).encode("utf-8")
            )
        )

    cov = coverage.coverage(
        source=[os.path.abspath("../")], config_file=".coveragerc", data_suffix=True
    )
    cov.start()

    import nose
    import sys

    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "../../")))

    try:
        nose.core.runmodule(
            modules_to_test,
            argv=sys.argv + ["--with-progressive", "--nologcapture"],
            exit=False,
        )  # log capture get's a ton of output from the pipegraph. enable if you need it
    except Exception as e:
        print("error")
        print(e)
        pass
    cov.stop()
    cov.save()
    # os.system('coverage combine')
    # os.system('coverage html -d covhtml')
    print("coverage report is in covhtml/index.html")



import multiprocessing

try:
    import Queue

    queue = Queue
except ImportError:
    import queue
q = multiprocessing.Queue()
q.put("Hello")
while True:
    try:
        q.get(timeout=5)
        break
    except (queue.Empty):
        print("timeout")
print("leaving - all OK")



"""
The MIT License (MIT)

Copyright (c) 2012, Florian Finkernagel <finkernagel@imt.uni-marburg.de>

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from twisted.internet.protocol import Protocol, Factory, ClientFactory
from twisted.internet.error import CannotListenError
from twisted.internet import reactor
import logging
import cPickle

import sys

try:
    port = int(sys.argv[1])
except IndexError:
    port = 5005

of = open("log.txt", "wb")


class Debug(Protocol):
    def dataReceived(self, data):
        global of
        try:
            if data == "please exit":
                print("End log because other logger requested access")
                of.write("End log because other logger requested access")
                reactor.stop()
            lr = logging.makeLogRecord(cPickle.loads(data[4:]))
            if lr.getMessage().find("New Pipegraph") != -1:
                of.close()
                of = open("log.txt", "wb")
            print("%s:%i: %s" % (lr.name, lr.lineno, lr.getMessage()))
            of.write("%s:%i: %s\n" % (lr.name, lr.lineno, lr.getMessage()))
            of.flush()
        except cPickle.UnpicklingError:
            print("a messages was missing")
        # self.transport.write(data, (host, port))


class Killer(Protocol):
    def connectionMade(self):
        print("sending please exit")
        self.transport.write("please exit")  # no need for address
        self.transport.loseConnection()

    def connectionLost(self, reason):
        print("Other logger apperantly exited, now trying to listen again in 2 seconds")
        reactor.callLater(2, start_listening)


class KillerFactory(ClientFactory):
    protocol = Killer


def start_listening():
    factory = Factory()
    factory.protocol = Debug

    def listening():
        print("now listening")

    factory.startFactory = listening
    reactor.listenTCP(port, factory)


try:
    start_listening()
except CannotListenError:
    print("trying to send kill signal to already running instance")
    d = reactor.connectTCP("localhost", port, KillerFactory())

# print "Going to listen on port %i, logging to %s" % (port, of.name)
reactor.run()



import os
import pytest
import pypipegraph as ppg
from .shared import assertRaises, write


def forget_job_status(invariant_status_filename=None):
    """Delete the job status file - usually only useful for testing"""
    if invariant_status_filename is None:
        invariant_status_filename = ppg.graph.invariant_status_filename_default
    try:
        os.unlink(invariant_status_filename)
    except OSError:
        pass


def destroy_global_pipegraph():
    """Free the current global pipegraph - usually only useful for testing"""
    ppg.util.global_pipegraph = None


@pytest.mark.usefixtures("new_pipegraph")
class TestSimple:
    def test_job_creation_before_pipegraph_creation_raises(self):
        destroy_global_pipegraph()
        with pytest.raises(ValueError):
            ppg.FileGeneratingJob("A", lambda: None)

    def test_run_pipegraph_without_pipegraph_raises(self):
        destroy_global_pipegraph()
        with pytest.raises(ValueError):
            ppg.run_pipegraph()

    def test_can_not_run_twice(self):

        ppg.new_pipegraph(dump_graph=False)
        ppg.run_pipegraph()
        try:
            ppg.run_pipegraph()
            assert False  # "Exception not correctly raised"
        except ValueError as e:
            print(e)
            assert "Each pipegraph may be run only once." in str(e)

    def test_can_not_add_jobs_after_run(self):

        ppg.new_pipegraph(dump_graph=False)
        ppg.run_pipegraph()
        try:
            ppg.FileGeneratingJob("out/A", lambda: write("out/A", "A"))
            assert False  # , "Exception not correctly raised")
        except ValueError as e:
            print(e)
            assert (
                "This pipegraph was already run. You need to create a new one for more jobs"
                in str(e)
            )

    def test_job_creation_after_pipegraph_run_raises(self):
        def inner():
            ppg.FileGeneratingJob("A", lambda: None)

        ppg.new_pipegraph(quiet=True, dump_graph=False)
        ppg.run_pipegraph()
        assertRaises(ValueError, inner)

    def test_run_may_be_called_only_once(self):
        ppg.new_pipegraph(quiet=True, dump_graph=False)
        ppg.run_pipegraph()

        def inner():
            ppg.run_pipegraph()

        assertRaises(ValueError, inner)

    def test_non_default_status_filename(self):
        try:
            forget_job_status("shu.dat")
            forget_job_status()
            ppg.new_pipegraph(
                quiet=True, invariant_status_filename="shu.dat", dump_graph=False
            )
            ppg.FileGeneratingJob("out/A", lambda: write("out/A", "A"))
            ppg.run_pipegraph()
            assert os.path.exists("shu.dat")
            assert not (os.path.exists(ppg.graph.invariant_status_filename_default))
        finally:
            forget_job_status("shu.dat")



"""
The MIT License (MIT)

Copyright (c) 2012, Florian Finkernagel <finkernagel@imt.uni-marburg.de>

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import pytest
import pypipegraph as ppg
from .shared import write, assertRaises, read, append, writeappend, Dummy

shu = None


@pytest.mark.usefixtures("new_pipegraph")
class TestJobGeneratingJob:
    def test_basic(self):
        def gen():
            ppg.FileGeneratingJob("out/A", lambda: write("out/A", "A"))
            ppg.FileGeneratingJob("out/B", lambda: write("out/B", "B"))
            ppg.FileGeneratingJob("out/C", lambda: write("out/C", "C"))

        ppg.JobGeneratingJob("genjob", gen)
        ppg.run_pipegraph()
        assert read("out/A") == "A"
        assert read("out/B") == "B"
        assert read("out/C") == "C"

    def test_raises_if_needs_more_cores_than_we_have(self):
        def gen():
            jobA = ppg.FileGeneratingJob("out/A", lambda: write("out/A", "A"))
            jobA.cores_needed = 20000

        ppg.JobGeneratingJob("genjob", gen)
        try:
            ppg.run_pipegraph()
            raise ValueError("should not be reached")
        except ppg.RuntimeError:
            pass
        assert not (os.path.exists("out/A"))  # since the gen job crashed
        jobGenerated = ppg.util.global_pipegraph.jobs["out/A"]
        assert jobGenerated.failed
        assert jobGenerated.error_reason == "Needed to much memory/cores"

    def test_raises_if_needs_more_ram_than_we_have(self):
        def gen():
            jobA = ppg.FileGeneratingJob("out/A", lambda: write("out/A", "A"))
            jobA.memory_needed = 1024 * 1024 * 1024 * 1024

        ppg.JobGeneratingJob("genjob", gen)
        try:
            ppg.run_pipegraph()
            raise ValueError("should not be reached")
        except ppg.RuntimeError:
            pass
        assert not (os.path.exists("out/A"))  # since the gen job crashed
        jobGenerated = ppg.util.global_pipegraph.jobs["out/A"]
        assert jobGenerated.failed
        assert jobGenerated.error_reason == "Needed to much memory/cores"

    def test_with_memory_needed(self):
        jobA = ppg.FileGeneratingJob("out/A", lambda: write("out/A", "A"))
        jobA.memory_needed = 1024
        ppg.run_pipegraph()
        assert os.path.exists("out/A")  # since the gen job crashed

    def test_injecting_multiple_stages(self):
        def gen():
            def genB():
                def genC():
                    ppg.FileGeneratingJob("out/D", lambda: write("out/D", "D"))

                ppg.JobGeneratingJob("C", genC)

            ppg.JobGeneratingJob("B", genB)

        ppg.JobGeneratingJob("A", gen)
        ppg.run_pipegraph()
        assert read("out/D") == "D"

    def test_generated_job_depending_on_each_other_one_of_them_is_Invariant(
        self, new_pipegraph
    ):
        # basic idea. You have jobgen A,
        # it not only creates filegenB, but also ParameterDependencyC that A depends on
        # does that work
        def gen():
            jobB = ppg.FileGeneratingJob("out/B", lambda: write("out/B", "B"))
            jobB.ignore_code_changes()
            jobC = ppg.ParameterInvariant("C", ("ccc",))
            jobB.depends_on(jobC)

        ppg.JobGeneratingJob("A", gen)
        ppg.run_pipegraph()
        assert read("out/B") == "B"

        new_pipegraph.new_pipegraph()

        def gen2():
            jobB = ppg.FileGeneratingJob("out/B", lambda: write("out/B", "C"))
            jobB.ignore_code_changes()
            jobC = ppg.ParameterInvariant("C", ("ccc",))
            jobB.depends_on(jobC)

        ppg.JobGeneratingJob("A", gen2)
        ppg.run_pipegraph()
        assert read("out/B") == "B"  # no rerun

        new_pipegraph.new_pipegraph()

        def gen3():
            jobB = ppg.FileGeneratingJob("out/B", lambda: write("out/B", "C"))
            jobB.ignore_code_changes()
            jobCX = ppg.ParameterInvariant("C", ("DDD",))
            jobB.depends_on(jobCX)

        ppg.JobGeneratingJob("A", gen3)
        ppg.run_pipegraph()
        assert read("out/B") == "C"  # did get rerun

    def test_generated_job_depending_on_job_that_cant_have_finished(
        self, new_pipegraph
    ):
        # basic idea. You have jobgen A, and filegen B.
        # filegenB depends on jobgenA.
        # jobGenA created C depends on filegenB
        # Perhaps add a filegen D that's independand of jobGenA, but C also deps on D
        def a():
            jobB = ppg.FileGeneratingJob("out/B", lambda: write("out/B", "B"))

            def genA():
                jobC = ppg.FileGeneratingJob("out/C", lambda: write("out/C", "C"))
                jobC.depends_on(jobB)

            jobA = ppg.JobGeneratingJob("A", genA)
            jobB.depends_on(jobA)
            ppg.run_pipegraph()
            assert read("out/B") == "B"
            assert read("out/C") == "C"

        def b():
            jobB = ppg.FileGeneratingJob("out/B", lambda: write("out/B", "B"))
            jobD = ppg.FileGeneratingJob("out/D", lambda: write("out/D", "D"))

            def genA():
                jobC = ppg.FileGeneratingJob("out/C", lambda: write("out/C", "C"))
                jobC.depends_on(jobB)
                jobC.depends_on(jobD)

            jobA = ppg.JobGeneratingJob("A", genA)
            jobB.depends_on(jobA)
            ppg.run_pipegraph()
            assert read("out/B") == "B"
            assert read("out/C") == "C"

        a()
        new_pipegraph.new_pipegraph()
        b()

    def test_generated_job_depending_on_each_other(self):
        # basic idea. You have jobgen A,
        # it not only creates filegenB, but also filegenC that depends on B
        # does that work
        def gen():
            jobB = ppg.FileGeneratingJob("out/B", lambda: write("out/B", "B"))
            jobC = ppg.FileGeneratingJob("out/C", lambda: write("out/C", read("out/B")))
            jobC.depends_on(jobB)

        ppg.JobGeneratingJob("A", gen)
        ppg.run_pipegraph()
        assert read("out/B") == "B"
        assert read("out/C") == "B"

    def test_generated_job_depending_on_each_other_one_of_them_is_loading(self):
        # basic idea. You have jobgen A,
        # it not only creates filegenB, but also DataloadingC that depends on B
        # does that work
        def gen():
            def load():
                global shu
                shu = "123"

            def do_write():
                global shu
                write("out/A", shu)

            dl = ppg.DataLoadingJob("dl", load)
            jobB = ppg.FileGeneratingJob("out/A", do_write)
            jobB.depends_on(dl)

        ppg.JobGeneratingJob("gen", gen)
        ppg.run_pipegraph()
        assert read("out/A") == "123"

    def test_passing_non_function(self):
        def inner():
            ppg.JobGeneratingJob("out/a", "shu")

        assertRaises(ValueError, inner)

    def test_passing_non_string_as_jobid(self):
        def inner():
            ppg.JobGeneratingJob(5, lambda: 1)

        assertRaises(TypeError, inner)

    def test_generated_jobs_that_can_not_run_right_away_because_of_dataloading_do_not_crash(
        self
    ):
        o = Dummy()
        existing_dl = ppg.AttributeLoadingJob("a", o, "a", lambda: "Ashu")

        def gen():
            new_dl = ppg.AttributeLoadingJob("b", o, "b", lambda: "Bshu")
            fg_a = ppg.FileGeneratingJob("out/C", lambda: write("out/C", o.a))
            fg_b = ppg.FileGeneratingJob("out/D", lambda: write("out/D", o.b))
            fg_a.depends_on(existing_dl)
            fg_b.depends_on(new_dl)

        ppg.JobGeneratingJob("E", gen)
        ppg.run_pipegraph()
        assert read("out/C") == "Ashu"
        assert read("out/D") == "Bshu"

    def test_filegen_invalidated_jobgen_created_filegen_later_also_invalidated(
        self, new_pipegraph
    ):
        a = ppg.FileGeneratingJob("out/A", lambda: writeappend("out/A", "out/Ac", "A"))
        p = ppg.ParameterInvariant("p", "p")
        a.depends_on(p)

        def gen():
            c = ppg.FileGeneratingJob(
                "out/C", lambda: writeappend("out/C", "out/Cx", "C")
            )
            c.depends_on(a)

        ppg.JobGeneratingJob("b", gen)
        ppg.run_pipegraph()
        assert read("out/A") == "A"
        assert read("out/Ac") == "A"
        assert read("out/C") == "C"
        assert read("out/Cx") == "C"
        new_pipegraph.new_pipegraph()

        a = ppg.FileGeneratingJob("out/A", lambda: writeappend("out/A", "out/Ac", "A"))
        p = ppg.ParameterInvariant("p", "p2")
        a.depends_on(p)
        ppg.JobGeneratingJob("b", gen)
        ppg.run_pipegraph()
        assert read("out/Ac") == "AA"
        assert read("out/Cx") == "CC"

    def test_raises_if_generating_within_dataload(self):
        ppg.util.global_pipegraph.quiet = False
        write_job = ppg.FileGeneratingJob("out/A", lambda: write("out/A", "aa"))

        def load():
            ppg.FileGeneratingJob("out/B", lambda: write("out/B", "aa"))

        dl = ppg.DataLoadingJob("load_data", load)
        write_job.depends_on(dl)
        with pytest.raises(ppg.RuntimeError):
            ppg.run_pipegraph()
        assert "Trying to add new jobs to running pipeline" in str(dl.exception)

    def test_ignored_if_generating_within_filegenerating(self):
        write_job = ppg.FileGeneratingJob("out/A", lambda: write("out/A", "aa"))

        def load():
            ppg.FileGeneratingJob("out/B", lambda: write("out/B", "aa"))
            write("out/C", "c")

        dl = ppg.FileGeneratingJob("out/C", load)
        write_job.depends_on(dl)
        ppg.run_pipegraph()
        assert read("out/C") == "c"

    def test_jobgenerating_is_not_dependency_injection(self):
        old = ppg.FileGeneratingJob("out/D", lambda: write("out/D", "D"))

        def gen():
            write("out/E", "E")
            p = ppg.FileGeneratingJob("out/C", lambda: write("out/C", "C"))
            old.depends_on(p)

        j = ppg.JobGeneratingJob("genjob", gen)
        with pytest.raises(ppg.RuntimeError):
            ppg.run_pipegraph()
        assert isinstance(j.exception, ppg.JobContractError)
        assert read("out/E") == "E"
        assert not os.path.exists("out/C")  # that job never makes it to the pipeline
        assert read("out/D") == "D"


@pytest.mark.usefixtures("new_pipegraph")
class TestDependencyInjectionJob:
    def test_basic(self, new_pipegraph):
        # TODO: there is a problem with this apporach. The AttributeLoadingJob
        # references different objects, since it get's pickled alongside with the method,
        # and depickled again, and then it's not the same object anymore,
        # so the FileGeneratingJob and the AttributeLoadingJob in this test
        # reference different objects.
        # I'm not sure how to handle this right now though.

        # I have an idea: Do JobGraphModifyingJobs in each slave, and send back just the
        # dependency data (and new job name).
        # that way, we can still execute on any slave, and all the pointers should be
        # right.
        new_pipegraph.new_pipegraph()

        o = Dummy()
        of = "out/A"

        def do_write():
            # logging.info("Accessing dummy (o) %i in pid %s" % (id(o), os.getpid()))
            write(of, o.A + o.B)

        job = ppg.FileGeneratingJob(of, do_write)

        def generate_deps():
            def load_a():
                # logging.info('executing load A')
                return "A"

            def load_b():
                # logging.info('executing load B')
                return "B"

            # logging.info("Creating dl on %i in pid %s" % (id(o), os.getpid()))
            dlA = ppg.AttributeLoadingJob("dlA", o, "A", load_a)
            # logging.info("created dlA")
            dlB = ppg.AttributeLoadingJob("dlB", o, "B", load_b)
            job.depends_on(dlA)
            job.depends_on(dlB)
            return [dlA, dlB]

        gen_job = ppg.DependencyInjectionJob("C", generate_deps)
        job.depends_on(gen_job)
        ppg.run_pipegraph()
        assert read(of) == "AB"

    def test_raises_on_non_dependend_job_injection(self):
        o = Dummy()
        of = "out/A"

        def do_write():
            write(of, o.A + o.B)

        job = ppg.FileGeneratingJob(of, do_write)
        jobD = ppg.FileGeneratingJob("out/D", lambda: write("out/D", "D"))

        def generate_deps():
            def load_a():
                return "A"

            def load_b():
                return "B"

            dlA = ppg.AttributeLoadingJob("dlA", o, "A", load_a)
            dlB = ppg.AttributeLoadingJob("dlB", o, "B", load_b)
            job.depends_on(dlA)
            jobD.depends_on(dlB)  # this line must raise

        gen_job = ppg.DependencyInjectionJob("C", generate_deps)
        job.depends_on(gen_job)
        with pytest.raises(ppg.RuntimeError):
            ppg.run_pipegraph()

        assert not (os.path.exists(of))  # since the gen job crashed
        assert os.path.exists(
            "out/D"
        )  # since it has no relation to the gen job actually...
        assert isinstance(gen_job.exception, ppg.JobContractError)
        assert "was not dependand on the " in str(gen_job.exception)

    def test_raises_on_non_dependend_job_injection2(self):
        o = Dummy()
        of = "out/A"

        def do_write():
            write(of, o.A + o.B)

        job = ppg.FileGeneratingJob(of, do_write)
        ppg.FileGeneratingJob("out/D", lambda: write("out/D", "D"))

        def generate_deps():
            def load_a():
                return "A"

            def load_b():
                return "B"

            dlA = ppg.AttributeLoadingJob("dlA", o, "A", load_a)
            ppg.AttributeLoadingJob("dlB", o, "B", load_b)
            job.depends_on(dlA)
            # let's not do anything with dlA

        gen_job = ppg.DependencyInjectionJob("C", generate_deps)
        job.depends_on(gen_job)
        with pytest.raises(ppg.RuntimeError):
            ppg.run_pipegraph()

        assert not (os.path.exists(of))  # since the gen job crashed
        assert os.path.exists(
            "out/D"
        )  # since it has no relation to the gen job actually...
        assert isinstance(gen_job.exception, ppg.JobContractError)
        assert "case 1" in str(gen_job.exception)

    def test_raises_on_non_dependend_job_injection2_can_be_ignored(self):
        o = Dummy()
        of = "out/A"

        def do_write():
            write(of, o.A)  # + o.B - but B is not in the dependency chain!

        job = ppg.FileGeneratingJob(of, do_write)
        ppg.FileGeneratingJob("out/D", lambda: write("out/D", "D"))

        def generate_deps():
            def load_a():
                return "A"

            def load_b():
                return "B"

            dlA = ppg.AttributeLoadingJob("dlA", o, "A", load_a)
            ppg.AttributeLoadingJob("dlB", o, "B", load_b)
            job.depends_on(dlA)
            # let's not do anything with dlA

        gen_job = ppg.DependencyInjectionJob(
            "C", generate_deps, check_for_dependency_injections=False
        )
        job.depends_on(gen_job)
        ppg.run_pipegraph()

        assert os.path.exists(of)  # since the gen job crashed

    def test_injecting_filegenerating_job(self):
        of = "out/A"

        def do_write():
            write(of, read("out/B"))

        job = ppg.FileGeneratingJob(of, do_write)

        def generate_dep():
            def write_B():
                write("out/B", "B")

            inner_job = ppg.FileGeneratingJob("out/B", write_B)
            job.depends_on(inner_job)

        job_gen = ppg.DependencyInjectionJob("gen_job", generate_dep)
        job.depends_on(job_gen)
        ppg.run_pipegraph()
        assert read("out/A") == "B"

    def test_passing_non_function(self):
        def inner():
            ppg.DependencyInjectionJob("out/a", "shu")

        assertRaises(ValueError, inner)

    def test_passing_non_string_as_jobid(self):
        def inner():
            ppg.DependencyInjectionJob(5, lambda: 1)

        assertRaises(TypeError, inner)

    def test_injecting_into_data_loading_does_not_retrigger(self, new_pipegraph):
        o = Dummy()

        def do_write():
            append("out/A", o.a + o.b)
            append("out/B", "X")

        def dl_a():
            o.a = "A"

        def do_run():
            of = "out/A"

            def inject():
                def dl_b():
                    o.b = "B"

                job_dl_b = ppg.DataLoadingJob("ob", dl_b)
                job_dl.depends_on(job_dl_b)

            job_fg = ppg.FileGeneratingJob(of, do_write)
            job_dl = ppg.DataLoadingJob("oa", dl_a)
            job_fg.depends_on(job_dl)
            job_inject = ppg.DependencyInjectionJob("inject", inject)
            job_dl.depends_on(job_inject)
            ppg.run_pipegraph()

        do_run()
        assert read("out/A") == "AB"
        assert read("out/B") == "X"
        new_pipegraph.new_pipegraph()
        do_run()
        assert read("out/A") == "AB"  # same data
        assert read("out/B") == "X"  # no rerun!
        # now let's test if a change triggers the rerun

        def do_run2():
            of = "out/A"

            def inject():
                def dl_b():
                    o.b = "C"  # so this dl has changed...

                job_dl_b = ppg.DataLoadingJob("ob", dl_b)
                job_dl.depends_on(job_dl_b)

            job_fg = ppg.FileGeneratingJob(of, do_write)
            job_dl = ppg.DataLoadingJob("oa", dl_a)
            job_fg.depends_on(job_dl)
            job_inject = ppg.DependencyInjectionJob("inject", inject)
            job_dl.depends_on(job_inject)
            ppg.run_pipegraph()

        new_pipegraph.new_pipegraph()
        do_run2()
        assert read("out/A") == "AC"  # same data
        assert read("out/B") == "XX"  # one rerun...



"""
The MIT License (MIT)

Copyright (c) 2012, Florian Finkernagel <finkernagel@imt.uni-marburg.de>

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import pytest
import pypipegraph as ppg
from .shared import assertRaises


@pytest.mark.usefixtures("new_pipegraph")
class TestUtils:
    def test_assert_uniqueness_simple(self):
        class Dummy:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self)

        Dummy("shu")

        def inner():
            Dummy("shu")

        assertRaises(ValueError, inner)

    def test_assert_uniqueness_ok(self):
        class Dummy:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self)

        Dummy("shu")
        Dummy("sha")

        def inner():
            Dummy("shu")

        assertRaises(ValueError, inner)

    def test_assert_uniqueness_ok_multi_classes(self):
        class Dummy:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self)

        class Dummy2:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self)

        Dummy("shu")
        Dummy2("shu")

        def inner():
            Dummy("shu")

        assertRaises(ValueError, inner)

    def test_assert_uniqueness_raises_slashes(self):
        class Dummy:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self)

        Dummy("shu")

        def inner():
            Dummy("shu/sha")

        assertRaises(ValueError, inner)

    def test_assert_uniqueness_raises_also_check(self):
        class Dummy:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self)

        class Dummy2:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self, also_check=Dummy)

        Dummy("shu")

        def inner():
            Dummy2("shu")

        assertRaises(ValueError, inner)

    def test_assert_uniqueness_raises_also_check_no_instance_of_second_class(self):
        class Dummy:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self)

        class Dummy2:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self, also_check=Dummy)

        # a = Dummy('shu')
        # does not raise of course...
        Dummy2("shu")

        def inner():
            Dummy2("shu")

        assertRaises(ValueError, inner)

    def test_assert_uniqueness_raises_also_check_list(self):
        class Dummy:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self)

        class Dummy2:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self, also_check=[Dummy])

        Dummy("shu")

        def inner():
            Dummy2("shu")

        assertRaises(ValueError, inner)

    def test_exception_on_run_without_previous_new_pipegraph(self):
        ppg.util.global_pipegraph = None
        with pytest.raises(ValueError):
            ppg.run_pipegraph()
